import logging
import json
import time
import os
import numpy as np
import dtlpy as dl

logger = logging.getLogger(name='dtlpy')

class VideoPlayer:
    """
    Video Player GUI.
    """

    def __init__(self, project_name=None, project_id=None,
                 dataset_name=None, dataset_id=None,
                 item_filepath=None, item_id=None):
        try:
            import tkinter
        except ImportError:
            logger.error(
                'Import Error! Cant import tkinter. Annotations operations will be limited. import manually and fix errors')
            raise

        # init tkinter window
        self.window = tkinter.Tk()
        self.window.title('Dataloop Player')

        self.video_source = None
        self.video_annotations = None
        self.local_annotations_filename = None
        self.labels = None
        self.project = None
        self.dataset = None
        self.item = None
        #############################
        # load video and annotation #
        #############################
        self.platform_params = {'project_name': project_name, 'project_id': project_id,
                                'dataset_name': dataset_name, 'dataset_id': dataset_id,
                                'item_filepath': item_filepath, 'item_id': item_id}
        self.from_platform()

        # load video
        self.photo = None
        self.vid = None
        self.load_video()

        # canvas to put frames on
        self.canvas = tkinter.Canvas(self.window, width=self.vid.width, height=self.vid.height)
        self.canvas.pack()
        self.canvas.bind('<Configure>', self._resize_image)

        ########################
        # create buttons panel #
        ########################
        self.window_width = self.vid.width
        self.window_height = self.vid.height
        self.buttons_window = tkinter.Toplevel(master=self.window)
        self.buttons_window.title('Controls')
        self.buttons_window.geometry()

        button_frame = tkinter.Frame(self.buttons_window)
        button_frame.grid(sticky="W", row=0, column=0)

        ###############
        # information #
        ###############
        txt_dataset_name = tkinter.Label(button_frame)
        txt_dataset_name.grid(sticky="W", row=2, column=0, columnspan=10)
        txt_dataset_name.configure(text='Dataset name: %s' % self.dataset.name)
        txt_item_name = tkinter.Label(button_frame)
        txt_item_name.grid(sticky="W", row=3, column=0, columnspan=10)
        txt_item_name.configure(text='Item name: %s' % self.item.name)
        ###############
        # Exit Button #
        ###############
        btn_exit = tkinter.Button(button_frame, text="Close", command=self.close).grid(sticky="W", row=17, column=0)
        ################
        # Play / Pause #
        ################
        self.btn_toggle_play = tkinter.Button(button_frame, text='Play ', command=self.toggle_play)
        self.btn_toggle_play.grid(sticky="W", row=4, column=0)
        ###################
        # Next prev frame #
        ###################
        btn_next_frame = tkinter.Button(button_frame, text="Next frame", command=self.next_frame).grid(sticky="W",
                                                                                                       row=4,
                                                                                                       column=2)
        btn_prev_frame = tkinter.Button(button_frame, text="Prev frame", command=self.prev_frame).grid(sticky="W",
                                                                                                       row=4,
                                                                                                       column=3)
        self.btn_toggle_frame_num = tkinter.Button(button_frame,
                                                   text="Hide frame number",
                                                   command=self.toggle_show_frame_number)
        self.btn_toggle_frame_num.grid(sticky="W", row=5, column=0, columnspan=10)
        self.btn_toggle_annotations = tkinter.Button(button_frame,
                                                     text="Hide annotations",
                                                     command=self.toggle_show_annotations)
        self.btn_toggle_annotations.grid(sticky="W", row=6, column=0, columnspan=10)
        self.btn_toggle_label = tkinter.Button(button_frame,
                                               text="Hide label",
                                               command=self.toggle_show_label)
        self.btn_toggle_label.grid(sticky="W", row=7, column=0, columnspan=10)
        #################
        # Export button #
        #################
        tkinter.Button(button_frame, text="Export video", command=self.export) \
            .grid(sticky="W", row=16, column=2, columnspan=2)
        # ################
        # # Apply button #
        # ################
        # btn_apply_offset_fix = tkinter.Button(button_frame, text="Apply offset fix", command=self.apply_offset_fix) \
        #     .grid(sticky="W", row=16, column=0, columnspan=2)
        btn_reset = tkinter.Button(button_frame, text="Reset", command=self.reset).grid(sticky="W", row=4, column=4)
        # #####################
        # # Annotation Offset #
        # #####################
        # tkinter.Label(button_frame, text="Annotation offset (in frames):") \
        #     .grid(sticky="W", row=13, column=0,
        #           columnspan=10)
        # self.annotations_offset_entry = tkinter.Entry(button_frame)
        # self.annotations_offset_entry.bind("<Return>", self.set_annotation_offset)
        # self.annotations_offset_entry.grid(sticky="W", row=14, column=0, columnspan=10)
        # self.annotation_offset_text = tkinter.Label(button_frame)
        # self.annotation_offset_text.grid(sticky="W", row=15, column=0, columnspan=10)
        ##############
        # Set Frames #
        ##############
        tkinter.Label(button_frame, text="Set frame (in frames):") \
            .grid(sticky="W", row=10, column=0, columnspan=10)
        self.current_frame_entry = tkinter.Entry(button_frame)
        self.current_frame_entry.bind("<Return>", self.set_current_frame)
        self.current_frame_entry.grid(sticky="W", row=11, column=0, columnspan=10)
        self.current_frame_text = tkinter.Label(button_frame)
        self.current_frame_text.grid(sticky="W", row=12, column=0, columnspan=10)
        #########################
        # Timestamp information #
        #########################
        self.frame_timestamp_text = tkinter.Label(button_frame)
        self.frame_timestamp_text.grid(sticky="W", row=8, column=0, columnspan=10)
        self.annotations_timestamp_text = tkinter.Label(button_frame)
        self.annotations_timestamp_text.grid(sticky="W", row=9, column=0, columnspan=10)

        ##############
        # Parameters #
        ##############
        self.delay = None
        self.annotations_timestamp = None
        # self.annotations_offset = None
        self.annotations_timestamp = None
        self.show_frame_num = True
        self.show_annotations = True
        self.show_label = True
        self.playing = False
        self.init_all_params()
        self.show_frame(0)

        ###############
        # Start video #
        ###############
        self.delay = int(1000 * 1 / self.vid.fps)
        self.update()
        button_frame.lift()
        self.window.mainloop()

    def _resize_image(self, event):
        """
        Resize frame to match window size
        :param event:
        :return:
        """
        self.window_width = event.width
        self.window_height = event.height

    def reset(self):
        """
        Reset video and annotation
        :return:
        """
        self.from_platform()
        self.load_video()
        self.init_all_params()
        self.show_frame(0)

    def export(self):
        """
        Create an annotated video saved to file
        :return:
        """

        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        try:
            from tkinter import ttk
        except ImportError:
            logger.error(
                'Import Error! Cant import ttk from tkinter. Annotations operations will be limited. import manually and fix errors')
            raise

        # start progress bar
        p, ext = os.path.splitext(self.video_source)
        output_filename = p + '_out.mp4'

        # read input video
        reader = cv2.VideoCapture(self.video_source)
        width = int(reader.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(reader.get(cv2.CAP_PROP_FRAME_HEIGHT))
        fps = reader.get(cv2.CAP_PROP_FPS)
        encoding = int(reader.get(cv2.CAP_PROP_FOURCC))
        n_frames = int(reader.get(cv2.CAP_PROP_FRAME_COUNT))
        writer = cv2.VideoWriter(output_filename, cv2.VideoWriter_fourcc(*'MP4V'), fps, (width, height))

        # create popup
        popup = tkinter.Toplevel()
        tkinter.Label(popup, text='Exporting to\n%s' % output_filename).grid(row=0, column=0)
        progress_var = tkinter.DoubleVar()
        progress_bar = ttk.Progressbar(popup, variable=progress_var, maximum=n_frames)
        progress_bar.grid(row=1, column=0)  # .pack(fill=tk.X, expand=1, side=tk.BOTTOM)
        popup.pack_slaves()

        i_frame = 0
        while reader.isOpened():
            popup.update()
            ret, frame = reader.read()
            if not ret:
                break
            # mark on frame
            annotations = [frame_annotation.get_annotation_by_frame(frame=i_frame)
                           for frame_annotation in self.video_annotations]

            for annotation in annotations:
                if annotation is None:
                    continue
                frame = annotation.show(image=frame,
                                        color=self.get_class_color(annotation.label))
                if self.show_label:
                    text = '%s-%s' % (annotation.label, ','.join(annotation.attributes))
                    frame = cv2.putText(frame,
                                        text=text,
                                        org=tuple([int(np.round(annotation.left)), int(np.round(annotation.top))]),
                                        color=(255, 0, 0),
                                        fontFace=cv2.FONT_HERSHEY_DUPLEX,
                                        fontScale=1,
                                        thickness=2)
            # write
            writer.write(frame)
            i_frame += 1
            progress_var.set(i_frame)
        reader.release()
        writer.release()
        popup.destroy()

    def load_video(self):
        """
        Load VideoCapture instance
        :return:
        """
        self.vid = VideoCapture(self.video_source)

    def init_all_params(self):
        """
        Reset all initial variables
        :return:
        """
        self.annotations_timestamp = 0
        # self.annotations_offset = 0
        # self.annotation_offset_text.configure(text='Current: %d' % self.annotations_offset)
        self.annotations_timestamp_text.configure(text='Annotation timestamp:\n %d' % self.annotations_timestamp)
        self.annotations_timestamp_text.grid(sticky="W", row=9, column=0, columnspan=10)
        # set text frames
        # self.annotations_offset_entry.delete(0, 'end')
        # self.annotations_offset_entry.insert(0, str(self.annotations_offset))
        self.current_frame_entry.delete(0, 'end')
        self.current_frame_entry.insert(0, str(self.vid.frame_number))

    def from_local(self):
        """
        Load local video and annotation
        :return:
        """
        if self.local_annotations_filename is not None:
            with open(self.local_annotations_filename, 'r') as f:
                data = json.load(f)
        self.video_annotations = dl.AnnotationCollection.from_json(data['annotations'])

    def from_platform(self):
        """
        Load video and annotations from platform
        :return:
        """
        project_name = self.platform_params['project_name']
        project_id = self.platform_params['project_id']
        dataset_name = self.platform_params['dataset_name']
        dataset_id = self.platform_params['dataset_id']
        item_filepath = self.platform_params['item_filepath']
        item_id = self.platform_params['item_id']

        # load remote item
        if dataset_id is None:
            self.project = dl.projects.get(project_name=project_name, project_id=project_id)
            if self.project is None:
                raise ValueError('Project doesnt exists. name: %s, id: %s' % (project_name, project_id))
            self.dataset = self.project.datasets.get(dataset_name=dataset_name, dataset_id=dataset_id)
        else:
            self.dataset = dl.datasets.get(dataset_id=dataset_id)
        if self.dataset is None:
            raise ValueError('Dataset doesnt exists. name: %s, id: %s' % (dataset_name, dataset_id))
        self.item = self.dataset.items.get(filepath=item_filepath, item_id=item_id)
        if self.item is None:
            raise ValueError('Item doesnt exists. name: %s, id: %s' % (item_filepath, item_id))
        self.labels = {label.tag: label.rgb for label in self.dataset.labels}
        _, ext = os.path.splitext(self.item.filename[1:])
        video_filename = os.path.join(self.dataset.__get_local_path__(), self.item.filename[1:])
        if not os.path.isdir(os.path.dirname(video_filename)):
            os.makedirs(os.path.dirname(video_filename))
        if not os.path.isfile(video_filename):
            self.item.download(local_path=os.path.dirname(video_filename), to_items_folder=False)
        self.video_source = video_filename
        self.video_annotations = self.item.annotations.list()

    def close(self):
        """
        Terminate window and application
        :return:
        """
        self.window.destroy()
        self.buttons_window.destroy()

    #
    # def set_annotation_offset(self, entry):
    #     """
    #     Set annotations offset to video start time
    #     :param entry:
    #     :return:
    #     """
    #     self.annotations_offset = int(self.annotations_offset_entry.get())
    #     self.annotation_offset_text.configure(text='Current: %d' % self.annotations_offset)

    def set_current_frame(self, entry):
        """
        Go to specific frame
        :param entry:
        :return:
        """
        input_frame = int(self.current_frame_entry.get())
        self.show_frame(frame_number=input_frame)

    def toggle_show_frame_number(self):
        """
        Show/hide frame number on frames
        :return:
        """
        if self.show_frame_num:
            self.show_frame_num = False
            self.btn_toggle_frame_num.config(text='Show frame num')
        else:
            self.show_frame_num = True
            self.btn_toggle_frame_num.config(text='Hide frame num')

    def toggle_show_annotations(self):
        """
        Show/hide annotations from frame
        :return:
        """
        if self.show_annotations:
            self.show_annotations = False
            self.btn_toggle_annotations.config(text='Show annotations')
        else:
            self.show_annotations = True
            self.btn_toggle_annotations.config(text='Hide annotations')

    def toggle_show_label(self):
        """
        Show/hide label per annotations
        :return:
        """
        if self.show_label:
            self.show_label = False
            self.btn_toggle_label.config(text='Show label')
        else:
            self.show_label = True
            self.btn_toggle_label.config(text='Hide label')

    def get_class_color(self, label):
        """
        Color of label
        :param label:
        :return:
        """
        if label not in self.labels:
            print('[WARNING] label not in dataset labels: %s' % label)
            return (255, 0, 0)
        color = self.labels[label]
        if isinstance(color, str):
            if color.startswith('rgb'):
                color = tuple(eval(color.lstrip('rgb')))
            elif color.startswith('#'):
                color = tuple(int(color.lstrip('#')[i:i + 2], 16) for i in (0, 2, 4))
            else:
                print('[WARNING] Unknown color scheme: %s' % color)
                color = (255, 0, 0)
        return color

    def get_annotations(self, frame):
        """
        Get all annotations of frame
        :param frame:
        :return:
        """
        # self.annotations_timestamp = (self.vid.frame_number + self.annotations_offset) / self.vid.fps
        self.annotations_timestamp = self.vid.frame_number / self.vid.fps
        frame = self.video_annotations.get_frame(frame_num=self.vid.frame_number).show(image=frame,
                                                                                       height=frame.shape[0],
                                                                                       width=frame.shape[1],
                                                                                       with_text=self.show_label)
        return frame

    def toggle_play(self):
        """
        Toggle play pause
        :return:
        """
        if self.playing:
            # need to pause
            self.playing = False
            self.btn_toggle_play.config(text='Play ')
        else:
            # need to play
            self.playing = True
            self.btn_toggle_play.config(text='Pause')

    def next_frame(self):
        """
        Get next frame
        :return:
        """
        self.show_frame()

    def prev_frame(self):
        """
        Get previous frame
        :return:
        """
        self.show_frame(self.vid.frame_number - 1)

    # def apply_offset_fix(self):
    #     """
    #     Apply offset to platform
    #     :return:
    #     """
    #     # delete annotations from platform
    #     for annotation in self.video_annotations.annotations:
    #         self.item.annotations.delete(annotations_id=annotation.id)
    #
    #     # apply annotations fix
    #     new_annotations = list()
    #     for annotation in self.video_annotations.annotations:
    #         new_annotation = annotation.copy()
    #         for i in range(len(new_annotation['metadata']['snapshots_'])):
    #             new_annotation['metadata']['snapshots_'][i]['startTime'] -= (self.annotations_offset / self.vid.fps)
    #             new_annotation['metadata']['snapshots_'][i]['frameNum'] = int(
    #                 new_annotation['metadata']['snapshots_'][i]['startTime'] * self.vid.fps)
    #         new_annotations.append(json.dumps(new_annotation))
    #
    #     # upload annotations
    #     self.item.annotation.upload(annotation=[new_annotations])

    def show_frame(self, frame_number=None):
        """
        Get a frame from the video source
        :param frame_number:
        :return:
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        try:
            import tkinter
        except ImportError:
            logger.error(
                'Import Error! Cant import tkinter. Annotations operations will be limited. import manually and fix errors')
            raise

        try:
            import PIL.ImageTk
            import PIL.Image
        except ImportError:
            logger.error(
                'Import Error! Cant import PIL.ImageTk/PIL.ImageTk. Annotations operations will be limited. '
                'import manually and fix errors')
            raise
        ret, frame = self.vid.get_frame(frame_number)
        if ret:
            if self.show_annotations:
                frame = self.get_annotations(frame)
            if self.show_frame_num:
                text = '%d - %d' % (self.vid.frame_number, np.round(self.annotations_timestamp * self.vid.fps))
                frame = cv2.putText(frame,
                                    text=text,
                                    org=(100, 100),
                                    color=(0, 0, 255),
                                    fontFace=cv2.FONT_HERSHEY_DUPLEX,
                                    fontScale=2,
                                    thickness=3)

            self.photo = PIL.ImageTk.PhotoImage(
                image=PIL.Image.fromarray(frame).resize((self.window_width, self.window_height)),
            master=self.canvas)
            self.canvas.create_image(0, 0, image=self.photo, anchor=tkinter.NW)
            # set timestamp
            self.current_frame_text.configure(text='Frame number:\n%d' % self.vid.frame_number)
            self.current_frame_text.grid(sticky="W", row=12, column=0, columnspan=10)
            millis = int(1000 * self.vid.frame_number / self.vid.fps)
            seconds = (millis / 1000) % 60
            minutes = int((millis / (1000 * 60)) % 60)
            hours = int((millis / (1000 * 60 * 60)) % 24)
            self.frame_timestamp_text.configure(
                text='Frame timestamp:\n{:02d}:{:02d}:{:.3f}'.format(hours, minutes, seconds))
            self.frame_timestamp_text.grid(sticky="W", row=8, column=0, columnspan=10)

    def update(self):
        """
        Playing the movie. Get frames at the FPS and show
        :return:
        """
        tic = time.time()
        if self.playing:
            self.show_frame()
        add_delay = np.maximum(1, self.delay - int(1000 * (time.time() - tic)))
        self.window.after(ms=add_delay, func=self.update)


class VideoCapture:
    """
    Video class using cv2 to play video
    """
    def __init__(self, source=0):
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise
        # Open the video source
        self.vid = cv2.VideoCapture(source)
        if not self.vid.isOpened():
            raise ValueError("Unable to open video source", source)

        # Get video source width and height
        self.width = self.vid.get(cv2.CAP_PROP_FRAME_WIDTH)
        self.height = self.vid.get(cv2.CAP_PROP_FRAME_HEIGHT)
        self.fps = self.vid.get(cv2.CAP_PROP_FPS)
        self.frame_number = self.vid.get(cv2.CAP_PROP_POS_FRAMES)

    def get_frame(self, frame_number=None):
        """
        get the frame from video
        :param frame_number:
        :return:
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise
        if self.vid.isOpened():
            if self.frame_number is None:
                self.frame_number = self.vid.get(cv2.CAP_PROP_POS_FRAMES)
            else:
                self.frame_number += 1
            if frame_number is not None:
                self.frame_number = frame_number
                self.vid.set(cv2.CAP_PROP_POS_FRAMES, frame_number)
            ret, frame = self.vid.read()

            if ret:
                # Return a boolean success flag and the current frame converted to BGR
                return ret, cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            else:
                return ret, None
        else:
            return False, None

    # Release the video source when the object is destroyed
    def __del__(self):
        if self.vid.isOpened():
            self.vid.release()



def main():
    """
    Detect and track (using model and some tracker) and upload annotation to platform
    :return:
    """
    import cv2
    import dtlpy as dl

    ##########################
    # Load model and tracker #
    ##########################
    # load your model for detection
    model = load_some_model()
    # load any tracking algorithm to track detected elements
    tracker = load_some_tracker()

    ##############
    # load video #
    ##############
    video_path = 'some/video/path'

    vid = cv2.VideoCapture(video_path)
    if not vid.isOpened():
        raise IOError("Couldn't open webcam or video")
    video_fps = vid.get(cv2.CAP_PROP_FPS)
    video_size = (int(vid.get(cv2.CAP_PROP_FRAME_WIDTH)),
                  int(vid.get(cv2.CAP_PROP_FRAME_HEIGHT)))
    video_frames = int(vid.get(cv2.CAP_PROP_FRAME_COUNT))

    ############
    # Platform #
    ###########
    # get the item from platform
    item = dl.projects.get(project_name='MyProject') \
        .datasets.get(dataset_name='MyDataset') \
        .items.get(filepath='/path/to/video.mp4')
    builder = item.annotations.builder()

    #######
    # Run #
    #######
    frame_num = 0
    while True:
        # get new frame from video
        return_value, frame = vid.read()
        if not return_value:
            break

        # get detection
        detections = model.predict(frame)

        # update tracker
        tracked_elements = tracker.update(detections, frame)

        # update annotations object
        for element in tracked_elements:
            # element.bb - format of the bounding box is 2 points in 1 array - [x_left, y_top, x_right, y_bottom])
            # tracking id of each element is in element.id. to keep the ids of the detected elements
            left, top, bottom, right = element.bb  # points bounding box annotation
            builder.add(annotation_definition=dl.Box(top=top,
                                                     left=left,
                                                     right=right,
                                                     bottom=bottom,
                                                     label=element.label),
                        object_id=element.id,
                        frame_num=frame_num)
        # increase frame number
        frame_num += 1
        if cv2.waitKey(1) & 0xFF == ord('q'):
            break

    ##################################
    # Upload annotations to platform #
    ##################################
    item.annotations.upload(builder.to_platform())



#! /usr/bin/python3
import subprocess
import traceback
import datetime
import logging
import shlex
import sys
import os
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit import prompt
import dtlpy as dlp
from dtlpy import exceptions
from dtlpy.dlp.cli_utilities import FileHistory, get_parser_tree, DlpCompleter
from dtlpy.dlp.parser import get_parser
from dtlpy.dlp.command_executor import CommandExecutor

dlp.client_api.is_cli = True

##########
# Logger #
##########
# set levels for CLI
logger = logging.getLogger(name='dtlpy')
handler_id = next((i for i, hand in enumerate(logger.handlers) if isinstance(hand, logging.StreamHandler)), 0)
logger.handlers[handler_id].setLevel(level=logging.INFO)
logger.propagate = False


def dlp_exit():
    print(datetime.datetime.now(datetime.timezone.utc))
    print("Goodbye ;)")
    sys.exit(0)


def main():
    try:
        parser = get_parser()
        keywords = get_parser_tree(parser=parser)
        args = parser.parse_args()
        history_file = os.path.join(os.getcwd(), ".history.txt")
        command_executor = CommandExecutor(dl=dlp, parser=parser)

        if args.version:
            logger.info("Dataloop SDK Version: {}".format(dlp.__version__))

        elif args.operation == "shell":
            #######################
            # Open Dataloop shell #
            #######################
            while True:
                text = prompt(u"dl>",
                              history=FileHistory(history_file),
                              auto_suggest=AutoSuggestFromHistory(),
                              completer=DlpCompleter(keywords=keywords, dlp=dlp))
                if text == '':
                    # in new line
                    continue

                try:
                    if text in ["-h", "--help"]:
                        text = "help"
                    text = shlex.split(text)
                    if text[0] not in keywords:
                        p = subprocess.Popen(text)
                        p.communicate()
                        continue
                    parser = get_parser()
                    args = parser.parse_args(text)
                    if args.operation == "exit":
                        dlp_exit()
                    else:
                        command_executor.run(args=args)
                except exceptions.TokenExpired:
                    print(datetime.datetime.now(datetime.timezone.utc))
                    print("[ERROR] token expired, please login.")
                    continue
                except SystemExit as e:
                    # exit
                    if e.code == 0:
                        if "-h" in text or "--help" in text:
                            continue
                        else:
                            sys.exit(0)
                    # error
                    else:
                        print(datetime.datetime.now(datetime.timezone.utc))
                        print('"{command}" is not a valid command'.format(command=text))
                        continue
                except Exception as e:
                    print(datetime.datetime.now(datetime.timezone.utc))
                    if hasattr(e, 'message'):
                        print(e.message)
                    else:
                        print(e.__str__())
                    continue

        else:
            ######################
            # Run single command #
            ######################
            try:
                command_executor.run(args=args)
                sys.exit(0)
            except exceptions.TokenExpired:
                print(datetime.datetime.now(datetime.timezone.utc))
                print("[ERROR] token expired, please login.")
                sys.exit(1)
            except Exception as e:
                print(datetime.datetime.now(datetime.timezone.utc))
                print(traceback.format_exc())
                print(e)
                sys.exit(1)
    except KeyboardInterrupt:
        dlp_exit()
    except EOFError:
        dlp_exit()
    except Exception:
        print(traceback.format_exc())
        dlp_exit()


if __name__ == "__main__":
    try:
        main()
    except Exception as err:
        print(datetime.datetime.now(datetime.timezone.utc))
        print("[ERROR]\t%s" % err)
    print("Dataloop.ai CLI. Type dlp --help for options")



import logging
from .. import entities, miscellaneous, exceptions
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Bots:
    """
    Bots Repository

    The Bots class allows the user to manage bots and their properties.
    """

    def __init__(self, client_api: ApiClient, project: entities.Project):
        self._client_api = client_api
        self._project = project

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.bots repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ###########
    # methods #
    ###########
    def list(self) -> miscellaneous.List[entities.Bot]:
        """
        Get a project or package bots list.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :return: List of Bots objects
        :rtype: list

        **Example**:

        .. code-block:: python

            bots_list = service.bots.list()

        """
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/projects/{}/bots'.format(self.project.id))

        if success:
            bots_json = response.json()
            pool = self._client_api.thread_pools(pool_name='entity.create')
            jobs = [None for _ in range(len(bots_json))]
            # return triggers list
            for i_bot, bot in enumerate(bots_json):
                jobs[i_bot] = pool.submit(entities.Bot._protected_from_json,
                                          **{'project': self.project,
                                             'bots': self,
                                             'client_api': self._client_api,
                                             '_json': bot})

            # get all results
            results = [j.result() for j in jobs]
            # log errors
            _ = [logger.warning(r[1]) for r in results if r[0] is False]
            # return good jobs
            bots = miscellaneous.List([r[1] for r in results if r[0] is True])
        else:
            logger.error('Platform error getting bots')
            raise exceptions.PlatformException(response)
        return bots

    def get(self,
            bot_email: str = None,
            bot_id: str = None,
            bot_name: str = None):
        """
        Get a Bot object.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str bot_email: get bot by email
        :param str bot_id: get bot by id
        :param str bot_name: get bot by name
        :return: Bot object
        :rtype: dtlpy.entities.bot.Bot

        **Example**:

        .. code-block:: python

            service.bots.get(bot_id='bot_id')

        """
        if bot_id is None:
            if bot_name is not None:
                bots = self.list()
                bot = [bot for bot in bots if bot.name == bot_name]

            elif bot_email is not None:
                bots = self.list()
                bot = [bot for bot in bots if bot.email == bot_email]
            else:
                raise exceptions.PlatformException('400', 'Must choose by "bot_id" or "bot_name"')
            if not bot:
                # list is empty
                raise exceptions.PlatformException('404', 'Bot not found. Name: {}'.format(bot_email))
                # project = None
            elif len(bot) > 1:
                # more than one matching project
                raise exceptions.PlatformException('404', 'More than one bot with same name. Please "get" by id')
            else:
                bot_id = bot[0].id

        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/projects/{}/bots/{}'.format(self.project.id,
                                                                                            bot_id))
        if success:
            bot = entities.Bot.from_json(_json=response.json(),
                                         project=self.project,
                                         bots=self, client_api=self._client_api)
            # verify input bot name and bot email are same as the given id
            if bot_name is not None and bot.name != bot_name:
                logger.warning(
                    "Mismatch found in bots.get: bot_name is different then bot.name: "
                    "{!r} != {!r}".format(
                        bot_name,
                        bot.name))
            if bot_email is not None and bot.email != bot_email:
                logger.warning(
                    "Mismatch found in bots.get: bot_email is different then bot.email: "
                    "{!r} != {!r}".format(
                        bot_email,
                        bot.email))
        else:
            raise exceptions.PlatformException(response)

        assert isinstance(bot, entities.Bot)
        return bot

    def delete(self,
               bot_id: str = None,
               bot_email: str = None):
        """
        Delete a Bot.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        You must provide at least ONE of the following params: bot_id, bot_email

        :param str bot_id: bot id to delete
        :param str bot_email: bot email to delete
        :return: True if successful
        :rtype: bool

        **Example**:

        .. code-block:: python

            service.bots.delete(bot_id='bot_id')

        """
        if bot_id is None:
            if bot_email is None:
                raise exceptions.PlatformException(error='400',
                                                   message='must input one of bot_id or bot_email to delete')
            bot = self.get(bot_email=bot_email)
            bot_id = bot.id
        success, response = self._client_api.gen_request(req_type='delete',
                                                         path='/projects/{}/bots/{}'.format(self.project.id,
                                                                                            bot_id))
        if not success:
            raise exceptions.PlatformException(response)
        logger.info('Bot {} deleted successfully'.format(bot_id))
        return True

    def create(self,
               name: str,
               return_credentials: bool = False):
        """
        Create a new Bot.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str name: bot name
        :param str return_credentials: True will return the password when created
        :return: Bot object
        :rtype: dtlpy.entities.bot.Bot

        **Example**:

        .. code-block:: python

            service.bots.delete(name='bot', return_credentials=False)
        """
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/projects/{}/bots'.format(self.project.id),
                                                         json_req={'name': name,
                                                                   'returnCredentials': return_credentials})
        if success:
            bot = entities.Bot.from_json(_json=response.json(),
                                         project=self.project,
                                         bots=self, client_api=self._client_api)
        else:
            raise exceptions.PlatformException(response)
        assert isinstance(bot, entities.Bot)
        return bot



import numpy as np
import pandas as pd
import logging
import datetime

from .. import entities

logger = logging.getLogger(name='dtlpy')


class Results:
    def __init__(self, matches, annotation_type):
        self.matches = matches
        self.annotation_type = annotation_type

    def to_df(self):
        return self.matches.to_df()

    def summary(self):
        df = self.matches.to_df()
        total_set_one = len(df['first_id'].dropna())
        total_set_two = len(df['second_id'].dropna())
        # each set unmatched is the number of Nones from the other set
        unmatched_set_one = df.shape[0] - total_set_two
        unmatched_set_two = df.shape[0] - total_set_one
        matched_set_one = total_set_one - unmatched_set_one
        matched_set_two = total_set_two - unmatched_set_two
        # sanity
        assert matched_set_one == matched_set_two, 'matched numbers are not the same'
        assert df['annotation_score'].shape[0] == (unmatched_set_one + unmatched_set_two + matched_set_one), \
            'mis-match number if scores and annotations'
        return {
            'annotation_type': self.annotation_type,
            'mean_annotations_scores': df['annotation_score'].mean(),
            'mean_attributes_scores': df['attribute_score'].mean(),
            'mean_labels_scores': df['label_score'].mean(),
            'n_annotations_set_one': total_set_one,
            'n_annotations_set_two': total_set_two,
            'n_annotations_total': total_set_one + total_set_two,
            'n_annotations_unmatched_set_one': unmatched_set_one,
            'n_annotations_unmatched_set_two': unmatched_set_two,
            'n_annotations_unmatched_total': unmatched_set_one + unmatched_set_two,
            'n_annotations_matched_total': matched_set_one,
            'precision': matched_set_one / (matched_set_one + unmatched_set_two),
            'recall': matched_set_one / (matched_set_one + unmatched_set_one)
        }


class Match:
    def __init__(self,
                 first_annotation_id, first_annotation_label, first_annotation_confidence,
                 second_annotation_id, second_annotation_label, second_annotation_confidence,
                 # defaults
                 annotation_score=0, attributes_score=0, geometry_score=0, label_score=0):
        """
        Save a match between two annotations with all relevant scores

        :param first_annotation_id:
        :param second_annotation_id:
        :param annotation_score:
        :param attributes_score:
        :param geometry_score:
        :param label_score:
        """
        self.first_annotation_id = first_annotation_id
        self.first_annotation_label = first_annotation_label
        self.first_annotation_confidence = first_annotation_confidence
        self.second_annotation_id = second_annotation_id
        self.second_annotation_label = second_annotation_label
        self.second_annotation_confidence = second_annotation_confidence
        self.annotation_score = annotation_score
        self.attributes_score = attributes_score
        # Replace the old annotation score
        self.geometry_score = geometry_score
        self.label_score = label_score

    def __repr__(self):
        return 'annotation: {:.2f}, attributes: {:.2f}, geometry: {:.2f}, label: {:.2f}'.format(
            self.annotation_score, self.attributes_score, self.geometry_score, self.label_score)


class Matches:
    def __init__(self):
        self.matches = list()
        self._annotations_raw_df = list()

    def __len__(self):
        return len(self.matches)

    def __repr__(self):
        return self.to_df().to_string()

    def to_df(self):
        results = list()
        for match in self.matches:
            results.append({
                'first_id': match.first_annotation_id,
                'first_label': match.first_annotation_label,
                'first_confidence': match.first_annotation_confidence,
                'second_id': match.second_annotation_id,
                'second_label': match.second_annotation_label,
                'second_confidence': match.second_annotation_confidence,
                'annotation_score': match.annotation_score,
                'attribute_score': match.attributes_score,
                'geometry_score': match.geometry_score,
                'label_score': match.label_score,
            })
        df = pd.DataFrame(results)
        return df

    def add(self, match: Match):
        self.matches.append(match)

    def validate(self):
        first = list()
        second = list()
        for match in self.matches:
            if match.first_annotation_id in first:
                raise ValueError('duplication for annotation id {!r} in FIRST set'.format(match.first_annotation_id))
            if match.first_annotation_id is not None:
                first.append(match.first_annotation_id)
            if match.second_annotation_id in second:
                raise ValueError('duplication for annotation id {!r} in SECOND set'.format(match.second_annotation_id))
            if match.second_annotation_id is not None:
                second.append(match.second_annotation_id)
        return True

    def find(self, annotation_id, loc='first'):
        for match in self.matches:
            if loc == 'first':
                if match.first_annotation_id == annotation_id:
                    return match
            elif loc == 'second':
                if match.second_annotation_id == annotation_id:
                    return match
        raise ValueError('could not find annotation id {!r} in {}'.format(annotation_id, loc))


######################
# Matching functions #
######################
class Matchers:

    @staticmethod
    def calculate_iou_box(pts1, pts2, config):
        """
        Measure the two list of points IoU
        :param pts1: ann.geo coordinates
        :param pts2: ann.geo coordinates
        :return: `float` how Intersection over Union of tho shapes
        """
        try:
            from shapely.geometry import Polygon
        except (ImportError, ModuleNotFoundError) as err:
            raise RuntimeError('dtlpy depends on external package. Please install ') from err
        if len(pts1) == 2:
            # regular box annotation (2 pts)
            pt1_left_top = [pts1[0][0], pts1[0][1]]
            pt1_right_top = [pts1[0][0], pts1[1][1]]
            pt1_right_bottom = [pts1[1][0], pts1[1][1]]
            pt1_left_bottom = [pts1[1][0], pts1[0][1]]
        else:
            # rotated box annotation (4 pts)
            pt1_left_top = pts1[0]
            pt1_right_top = pts1[3]
            pt1_left_bottom = pts1[1]
            pt1_right_bottom = pts1[2]

        poly_1 = Polygon([pt1_left_top,
                          pt1_right_top,
                          pt1_right_bottom,
                          pt1_left_bottom])

        if len(pts2) == 2:
            # regular box annotation (2 pts)
            pt2_left_top = [pts2[0][0], pts2[0][1]]
            pt2_right_top = [pts2[0][0], pts2[1][1]]
            pt2_right_bottom = [pts2[1][0], pts2[1][1]]
            pt2_left_bottom = [pts2[1][0], pts2[0][1]]
        else:
            # rotated box annotation (4 pts)
            pt2_left_top = pts2[0]
            pt2_right_top = pts2[3]
            pt2_left_bottom = pts2[1]
            pt2_right_bottom = pts2[2]

        poly_2 = Polygon([pt2_left_top,
                          pt2_right_top,
                          pt2_right_bottom,
                          pt2_left_bottom])
        iou = poly_1.intersection(poly_2).area / poly_1.union(poly_2).area
        return iou

    @staticmethod
    def calculate_iou_classification(pts1, pts2, config):
        """
        Measure the two list of points IoU
        :param pts1: ann.geo coordinates
        :param pts2: ann.geo coordinates
        :return: `float` how Intersection over Union of tho shapes
        """
        return 1

    @staticmethod
    def calculate_iou_polygon(pts1, pts2, config):
        try:
            # from shapely.geometry import Polygon
            import cv2
        except (ImportError, ModuleNotFoundError) as err:
            raise RuntimeError('dtlpy depends on external package. Please install ') from err
        # # using shapley
        # poly_1 = Polygon(pts1)
        # poly_2 = Polygon(pts2)
        # iou = poly_1.intersection(poly_2).area / poly_1.union(poly_2).area

        # # using opencv
        width = int(np.ceil(np.max(np.concatenate((pts1[:, 0], pts2[:, 0]))))) + 10
        height = int(np.ceil(np.max(np.concatenate((pts1[:, 1], pts2[:, 1]))))) + 10
        mask1 = np.zeros((height, width))
        mask2 = np.zeros((height, width))
        mask1 = cv2.drawContours(
            image=mask1,
            contours=[pts1.round().astype(int)],
            contourIdx=-1,
            color=1,
            thickness=-1,
        )
        mask2 = cv2.drawContours(
            image=mask2,
            contours=[pts2.round().astype(int)],
            contourIdx=-1,
            color=1,
            thickness=-1,
        )
        iou = np.sum((mask1 + mask2) == 2) / np.sum((mask1 + mask2) > 0)
        if np.sum((mask1 + mask2) > 2):
            assert False
        return iou

    @staticmethod
    def calculate_iou_semantic(mask1, mask2, config):
        joint_mask = mask1 + mask2
        return np.sum(np.sum(joint_mask == 2) / np.sum(joint_mask > 0))

    @staticmethod
    def calculate_iou_point(pt1, pt2, config):
        """
        pt is [x,y]
        normalizing  to score  between [0, 1] -> 1 is the exact match
        if same point score is 1
        at about 20 pix distance score is about 0.5, 100 goes to 0
        :param pt1:
        :param pt2:
        :return:
        """
        """
        x = np.arange(int(diag))
        y = np.exp(-1 / diag * 20 * x)
        plt.figure()
        plt.plot(x, y)
        """
        height = config.get('height', 500)
        width = config.get('width', 500)
        diag = np.sqrt(height ** 2 + width ** 2)
        # 20% of the image diagonal tolerance (empirically). need to
        return np.exp(-1 / diag * 20 * np.linalg.norm(np.asarray(pt1) - np.asarray(pt2)))

    @staticmethod
    def match_attributes(attributes1, attributes2):
        """
        Returns IoU of the attributes. if both are empty - its a prefect match (returns 1)
        0: no matching
        1: perfect attributes match
        """
        if type(attributes1) is not type(attributes2):
            logger.warning('attributes are not same type: {}, {}'.format(type(attributes1), type(attributes2)))
            return 0

        if attributes1 is None and attributes2 is None:
            return 1

        if isinstance(attributes1, dict) and isinstance(attributes2, dict):
            # convert to list
            attributes1 = ['{}-{}'.format(key, val) for key, val in attributes1.items()]
            attributes2 = ['{}-{}'.format(key, val) for key, val in attributes2.items()]

        intersection = set(attributes1).intersection(set(attributes2))
        union = set(attributes1).union(attributes2)
        if len(union) == 0:
            # if there is no union - there are no attributes at all
            return 1
        return len(intersection) / len(union)

    @staticmethod
    def match_labels(label1, label2):
        """
        Returns 1 in one of the labels in substring of the second
        """
        return int(label1 in label2 or label2 in label1)

    @staticmethod
    def general_match(matches: Matches,
                      first_set: entities.AnnotationCollection,
                      second_set: entities.AnnotationCollection,
                      match_type,
                      match_threshold: float,
                      ignore_attributes=False,
                      ignore_labels=False):
        """

        :param matches:
        :param first_set:
        :param second_set:
        :param match_type:
        :param match_threshold:
        :param ignore_attributes:
        :param ignore_labels:
        :return:
        """
        annotation_type_to_func = {
            entities.AnnotationType.BOX: Matchers.calculate_iou_box,
            entities.AnnotationType.CLASSIFICATION: Matchers.calculate_iou_classification,
            entities.AnnotationType.SEGMENTATION: Matchers.calculate_iou_semantic,
            entities.AnnotationType.POLYGON: Matchers.calculate_iou_polygon,
            entities.AnnotationType.POINT: Matchers.calculate_iou_point,
        }
        df = pd.DataFrame(data=-1 * np.ones((len(second_set), len(first_set))),
                          columns=[a.id for a in first_set],
                          index=[a.id for a in second_set])
        for annotation_one in first_set:
            for annotation_two in second_set:
                if match_type not in annotation_type_to_func:
                    raise ValueError('unsupported type: {}'.format(match_type))
                if df[annotation_one.id][annotation_two.id] == -1:
                    try:
                        config = {'height': annotation_one._item.height if annotation_one._item is not None else 500,
                                  'width': annotation_one._item.width if annotation_one._item is not None else 500}
                        df[annotation_one.id][annotation_two.id] = annotation_type_to_func[match_type](
                            annotation_one.geo,
                            annotation_two.geo,
                            config)
                    except ZeroDivisionError:
                        logger.warning(
                            'Found annotations with area=0!: annotations ids: {!r}, {!r}'.format(annotation_one.id,
                                                                                                 annotation_two.id))
                        df[annotation_one.id][annotation_two.id] = 0
        # for debug - save the annotations scoring matrix
        matches._annotations_raw_df.append(df.copy())

        # go over all matches
        while True:
            # take max IoU score, list the match and remove annotations' ids from columns and rows
            # keep doing that until no more matches or lower than match threshold
            max_cell = df.max().max()
            if max_cell < match_threshold or np.isnan(max_cell):
                break
            row_index, col_index = np.where(df == max_cell)
            row_index = row_index[0]
            col_index = col_index[0]
            first_annotation_id = df.columns[col_index]
            second_annotation_id = df.index[row_index]
            first_annotation = [a for a in first_set if a.id == first_annotation_id][0]
            second_annotation = [a for a in second_set if a.id == second_annotation_id][0]
            geometry_score = df.iloc[row_index, col_index]
            labels_score = Matchers.match_labels(label1=first_annotation.label,
                                                 label2=second_annotation.label)
            attribute_score = Matchers.match_attributes(attributes1=first_annotation.attributes,
                                                        attributes2=second_annotation.attributes)

            # TODO use ignores for final score
            annotation_score = (geometry_score + attribute_score + labels_score) / 3
            matches.add(Match(first_annotation_id=first_annotation_id,
                              first_annotation_label=first_annotation.label,
                              first_annotation_confidence=
                              first_annotation.metadata.get('user', dict()).get('model', dict()).get('confidence', 1),
                              second_annotation_id=second_annotation_id,
                              second_annotation_label=second_annotation.label,
                              second_annotation_confidence=
                              second_annotation.metadata.get('user', dict()).get('model', dict()).get('confidence', 1),
                              geometry_score=geometry_score,
                              annotation_score=annotation_score,
                              label_score=labels_score,
                              attributes_score=attribute_score))
            df.drop(index=second_annotation_id, inplace=True)
            df.drop(columns=first_annotation_id, inplace=True)
        # add un-matched
        for second_id in df.index:
            second_annotation = [a for a in second_set if a.id == second_id][0]
            matches.add(match=Match(first_annotation_id=None,
                                    first_annotation_label=None,
                                    first_annotation_confidence=None,
                                    second_annotation_id=second_id,
                                    second_annotation_label=second_annotation.label,
                                    second_annotation_confidence=
                                    second_annotation.metadata.get('user', dict()).get('model', dict()).get(
                                        'confidence', 1),
                                    ))
        for first_id in df.columns:
            first_annotation = [a for a in first_set if a.id == first_id][0]
            matches.add(match=Match(first_annotation_id=first_id,
                                    first_annotation_label=first_annotation.label,
                                    first_annotation_confidence=
                                    first_annotation.metadata.get('user', dict()).get('model', dict()).get('confidence',
                                                                                                           1),
                                    second_annotation_id=None,
                                    second_annotation_label=None,
                                    second_annotation_confidence=None))
        return matches


def item_annotation_duration(item: entities.Item = None,
                             dataset: entities.Dataset = None,
                             project: entities.Project = None,
                             task: entities.Task = None,
                             assignment: entities.Assignment = None):
    if all(ent is None for ent in [item, dataset, project, assignment, task]):
        raise ValueError('At least one input to annotation duration must not be None')
    query = {
        "startTime": 0,
        "context": {
            "accountId": [],
            "orgId": [],
            "projectId": [],
            "datasetId": [],
            "taskId": [],
            "assignmentId": [],
            "itemId": [],
            "userId": [],
            "serviceId": [],
            "podId": [],
        },
        "measures": [
            {
                "measureType": "itemAnnotationDuration",
                "pageSize": 1000,
                "page": 0,
            },
        ]
    }
    # add context for analytics
    created_at = list()
    if item is not None:
        query['context']['itemId'].append(item.id)
        created_at.append(int(1000 * datetime.datetime.fromisoformat(item.created_at[:-1]).timestamp()))
    if task is not None:
        query['context']['taskId'].append(task.id)
        created_at.append(int(1000 * datetime.datetime.fromisoformat(task.created_at[:-1]).timestamp()))
    if dataset is not None:
        query['context']['datasetId'].append(dataset.id)
        created_at.append(int(1000 * datetime.datetime.fromisoformat(dataset.created_at[:-1]).timestamp()))
    if assignment is not None:
        query['context']['assignmentId'].append(assignment.id)
        # assignment doesnt have "created_at" attribute
    query['startTime'] = int(np.min(created_at))
    raw = project.analytics.get_samples(query=query, return_field=None, return_raw=True)
    res = {row['itemId']: row['duration'] for row in raw[0]['response']}
    if item.id not in res:
        total_time_s = 0
    else:
        total_time_s = res[item.id] / 1000
    return total_time_s



import numpy as np
import time

from . import Box
from ...services.api_client import client as api_client


class Note(Box):
    """
        Note annotation object
    """

    def __init__(
            self,
            left,
            top,
            right,
            bottom,
            label,
            attributes=None,
            messages=None,
            status='issue',
            assignee=None,
            create_time=None,
            creator=None,
            description=None
    ):
        super(Note, self).__init__(
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            label=label,
            attributes=attributes,
            description=description
        )
        self.type = "note"
        if messages is None:
            messages = []
        if not isinstance(messages, list):
            messages = [messages]
        for msg_index in range(len(messages)):
            if not isinstance(messages[msg_index], Message):
                messages[msg_index] = Message(body=messages[msg_index])
        self.messages = messages
        self.status = status
        self.create_time = create_time
        self.creator = creator
        if self.creator is None:
            self.creator = api_client.info()['user_email']
        self.assignee = assignee
        if self.assignee is None:
            self.assignee = self.creator

    def to_coordinates(self, color):
        box = super(Note, self).to_coordinates(color=color)
        note = {
            'messages': [msg.to_json() for msg in self.messages],
            'status': self.status,
            'createTime': self.create_time,
            'creator': self.creator,
            'assignee': self.assignee
        }
        coordinates = {
            'box': box,
            'note': note
        }

        return coordinates

    @staticmethod
    def from_coordinates(coordinates):
        return Box.from_coordinates(coordinates['box'])

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            geo = cls.from_coordinates(_json["coordinates"])
            note_data = _json["coordinates"].get('note', dict())
        elif "data" in _json:
            geo = cls.from_coordinates(_json["data"])
            note_data = _json["data"].get('note', dict())
        else:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))

        left = np.min(geo[:, 0])
        top = np.min(geo[:, 1])
        right = np.max(geo[:, 0])
        bottom = np.max(geo[:, 1])
        messages = [Message.from_json(msg) for msg in note_data.get('messages', list())]

        return cls(
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            label=_json["label"],
            attributes=_json.get("attributes", None),
            messages=messages,
            status=note_data.get('status', 'open'),
            creator=note_data.get('creator', 'me'),
            assignee=note_data.get('assignee', 'me'),
            create_time=note_data.get('createTime', 0),
        )

    def add_message(self, body: str = None):
        self.messages.append(Message(body=body))


class Message:
    """
    Note message object
    """

    def __init__(self, msg_id: str = None, creator: str = None, msg_time=None, body: str = None):
        self.id = msg_id
        self.time = msg_time if msg_time is not None else int(time.time() * 1000)
        self.body = body
        self.creator = creator
        if self.creator is None:
            self.creator = api_client.info()['user_email']

    def to_json(self):
        _json = {
            "id": self.id,
            "creator": self.creator,
            "time": self.time,
            "body": self.body
        }
        return _json

    @staticmethod
    def from_json(_json):
        return Message(
            msg_id=_json.get('id', None),
            msg_time=_json.get('time', None),
            body=_json.get('body', None),
            creator=_json.get('creator', None)
        )



import dtlpy as dl
import logging

logger = logging.getLogger(name='dtlpy')


class ServiceRunner(dl.BaseServiceRunner):
    """
    Package runner class

    """

    def __init__(self):
        """
        Init package attributes here

        :return:
        """

    def first_method(self, item, progress=None):
        """
        n this example we upload an annotation to item

        :param progress: Use this to update the progress of your package
        :return:
        """
        # these lines can be removed
        assert isinstance(progress, dl.Progress)
        progress.update(status='inProgress', progress=0)
        print('Item received! filename: {}'.format(item.filename))
        builder = item.annotations.builder()
        builder.add(dl.Classification(label='from_function'))
        item.annotations.upload(builder)
        print('Annotation uploaded!')

    def second_method(self, item, progress=None):
        """
        Write your main package service here

        :param progress: Use this to update the progress of your package
        :return:
        """
        # these lines can be removed
        assert isinstance(progress, dl.Progress)
        assert isinstance(item, dl.Item)


if __name__ == "__main__":
    """
    Run this main to locally debug your package
    """
    dl.packages.test_local_package()



import dtlpy as dl
import logging
import shutil
import tempfile
import os
import zipfile
import datetime

logger = logging.getLogger(name='dtlpy')


class ServiceRunner(dl.BaseServiceRunner):
    """
    Package runner class

    """

    def __init__(self):
        """
        Init package attributes here

        :return:
        """

    def run(self, dataset, query=None, progress=None):
        """
        In this example we create a task for annotated items in dataset

        :param query: Dictionary
        :param dataset: dl.Dataset
        :param progress: Use this to update the progress of your package
        :return:
        """
        local_path = tempfile.mkdtemp()

        try:

            converted_folder = os.path.join(local_path, dataset.name)
            filters = None
            if query is not None:
                filters = dl.Filters(resource=dl.FiltersResource.ITEM, custom_filter=query)

            pages = dataset.items.list(filters=filters)

            for i_page, page in enumerate(pages):
                for item in page:
                    self._convert_single_item(item=item, local_path=converted_folder)
                progress.update(status='inProgress', progress=(i_page + 1) / pages.total_pages_count)

            zip_filename = os.path.join(local_path,
                                        '{}_{}.zip'.format(dataset.name, int(datetime.datetime.now().timestamp())))
            self._zip_directory(zip_filename=zip_filename, directory=converted_folder)

            zip_item = dataset.items.upload(local_path=zip_filename,
                                            remote_path='/my_converted_annotations',
                                            overwrite=True)

            return zip_item.id

        except Exception:
            # implement exception handling
            pass
        finally:
            shutil.rmtree(local_path)

    def _convert_single_item(self, item, local_path):
        """
        Implement single item conversion here and save converted files to local path
        :param item:
        :return:
        """
        pass

    @staticmethod
    def _zip_directory(zip_filename, directory):
        """
        Method to zip a directory
        :param zip_filename:
        :param directory:
        :return:
        """
        zip_file = zipfile.ZipFile(zip_filename, 'w', zipfile.ZIP_DEFLATED)
        try:
            for root, dirs, files in os.walk(directory):
                for file in files:
                    filepath = os.path.join(root, file)
                    zip_file.write(filepath, arcname=os.path.relpath(filepath, directory))
        finally:
            zip_file.close()


if __name__ == "__main__":
    """
    Run this main to locally debug your package
    """
    dl.packages.test_local_package()



import logging
import numpy as np
import warnings

logger = logging.getLogger(name='dtlpy')


class BaseAnnotationDefinition:
    def __init__(self, description=None, attributes=None):
        self.description = description
        self._top = 0
        self._left = 0
        self._bottom = 0
        self._right = 0
        self._annotation = None
        if attributes and not isinstance(attributes, dict):
            raise TypeError('attributes should be a dictionary')
        self._attributes = attributes or {}

    @property
    def attributes(self):
        return self._attributes

    @attributes.setter
    def attributes(self, v):
        if v and not isinstance(v, dict):
            raise TypeError('attributes should be a dictionary')
        self._attributes = v

    @property
    def top(self):
        return self._top

    @top.setter
    def top(self, v):
        self._top = v

    @property
    def left(self):
        return self._left

    @left.setter
    def left(self, v):
        self._left = v

    @property
    def bottom(self):
        return self._bottom

    @bottom.setter
    def bottom(self, v):
        self._bottom = v

    @property
    def right(self):
        return self._right

    @right.setter
    def right(self, v):
        self._right = v

    @property
    def height(self):
        return np.round(self.bottom - self.top)

    @property
    def width(self):
        return np.round(self.right - self.left)

    @staticmethod
    def add_text_to_image(image, annotation):
        """
        :param image:
        :param annotation:
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        text = '{label}-{attributes}'.format(label=annotation.label, attributes=','.join(annotation.attributes))
        top = annotation.top
        left = annotation.left
        if top == 0:
            top = image.shape[0] / 10
        if left == 0:
            left = image.shape[1] / 10
        return cv2.putText(img=image,
                           text=text,
                           org=tuple([int(np.round(left)), int(np.round(top))]),
                           color=(255, 0, 0),
                           fontFace=cv2.FONT_HERSHEY_DUPLEX,
                           fontScale=1,
                           thickness=2)

    @property
    def logger(self):
        return logger



from .. import entities, exceptions
from ..services.api_client import ApiClient
from typing import List


class Collections:
    def __init__(self,
                 client_api: ApiClient,
                 item: entities.Item = None,
                 dataset: entities.Dataset = None
                 ):
        self._client_api = client_api
        self._dataset = dataset
        self._item = item

    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None:
            raise ValueError("Must set dataset for this action.")
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: entities.Dataset):
        self._dataset = dataset

    @property
    def item(self) -> entities.Item:
        if self._item is None:
            raise ValueError("Must set item for this action.")
        return self._item

    @item.setter
    def item(self, item: entities.Item):
        self._item = item

    def create(self, name: str) -> entities.Collection:
        """
        Creates a new collection in the dataset.

        :param name: The name of the new collection.
        :return: The created collection details.
        """
        self.validate_max_collections()
        self.validate_collection_name(name)
        payload = {"name": name}
        success, response = self._client_api.gen_request(
            req_type="post", path=f"/datasets/{self.dataset.id}/items/collections", json_req=payload
        )
        if success:
            collection_json = self._single_collection(data=response.json(), name=name)
            return entities.Collection.from_json(client_api=self._client_api, _json=collection_json)
        else:
            raise exceptions.PlatformException(response)

    def update(self, collection_name: str, new_name: str) -> entities.Collection:
        """
        Updates the name of an existing collection.

        :param collection_id: The ID of the collection to update.
        :param new_name: The new name for the collection.
        :return: The updated collection details.
        """
        self.validate_collection_name(new_name)
        payload = {"name": new_name}
        success, response = self._client_api.gen_request(
            req_type="patch", path=f"/datasets/{self.dataset.id}/items/collections/{collection_name}", json_req=payload
        )
        if success:
            collection_json = self._single_collection(data=response.json(), name=new_name)
            return entities.Collection.from_json(client_api=self._client_api, _json=collection_json)
        else:
            raise exceptions.PlatformException(response)

    def delete(self, collection_name: str) -> bool:
        """
        Deletes a collection from the dataset.

        :param collection_name: The name of the collection to delete.
        """
        success, response = self._client_api.gen_request(
            req_type="delete", path=f"/datasets/{self.dataset.id}/items/collections/{collection_name}"
        )
        if success:
            # Wait for the split operation to complete
            command = entities.Command.from_json(_json=response.json(),
                                                 client_api=self._client_api)
            command.wait()
            return True
        else:
            raise exceptions.PlatformException(response)

    def clone(self, collection_name: str) -> entities.Collection:
        """
        Clones an existing collection, creating a new one with a unique name.

        :param collection_name: The name of the collection to clone.
        :return: The cloned collection details as a dictionary.
        """
        self.validate_max_collections()
        collections = self.list_all_collections()
        original_collection = next((c for c in collections if c["name"] == collection_name), None)

        if not original_collection:
            raise ValueError(f"Collection with name '{collection_name}' not found.")

        source_name = original_collection["name"]
        num = 0
        clone_name = ""
        while True:
            num += 1
            clone_name = f"{source_name}-clone-{num}"
            if not any(c["name"] == clone_name for c in collections):  # Use c["name"] for comparison
                break

        # Create the cloned collection
        cloned_collection = self.create(name=clone_name)
        filters = entities.Filters()
        filters.add(field=f'metadata.system.collections.{original_collection["key"]}', values=True)
        self.assign(collections=[cloned_collection.name],
                    filters=filters)
        return cloned_collection

    def list_all_collections(self) -> List[dict]:
        """
        Retrieves all collections in the dataset.

        :return: A list of collections in the dataset.
        """
        success, response = self._client_api.gen_request(
            req_type="GET", path=f"/datasets/{self.dataset.id}/items/collections"
        )
        if success:
            data = response.json()
            return self._list_collections(data)
        else:
            raise exceptions.PlatformException(response)

    def validate_collection_name(self, name: str):
        """
        Validate that the collection name is unique.

        :param name: The name of the collection to validate.
        :raises ValueError: If a collection with the same name already exists.
        """
        collections = self.list_all_collections()
        if any(c["name"] == name for c in collections):
            raise ValueError(f"A collection with the name '{name}' already exists.")

    def validate_max_collections(self) -> None:
        """
        Validates that the dataset has not exceeded the maximum allowed collections.

        :raises ValueError: If the dataset has 10 or more collections.
        """
        collections = self.list_all_collections()
        if len(collections) >= 10:
            raise ValueError("The dataset already has the maximum number of collections (10).")

    def list_missing_collections(self) -> entities.PagedEntities:
        """
        List all items in the dataset that are not assigned to any collection.

        :return: A list of items that are not part of any collection.
        """
        if self._dataset is None:
            raise ValueError("Must set dataset for this action.")
        filters = entities.Filters(method=entities.FiltersMethod.AND)  # Use AND method for all conditions
        collection_fields = [
            "collections0",
            "collections1",
            "collections2",
            "collections3",
            "collections4",
            "collections5",
            "collections6",
            "collections7",
            "collections8",
            "collections9",
        ]

        # Add each field to the filter with a value of False
        for field in collection_fields:
            filters.add(field=field, values=False, method=entities.FiltersMethod.AND,
                        operator=entities.FiltersOperations.EXISTS)
        return self._dataset.items.list(filters=filters)

    def list_unassigned_items(self) -> list:
        """
        List unassigned items in a dataset (items where all collection fields are false).

        :return: List of unassigned item IDs
        :rtype: list
        """
        missing_ids = []
        pages = self.list_missing_collections()
        for page in pages:
            for item in page:
                # Items that pass filters mean all collections are false
                missing_ids.append(item.id)

        return missing_ids

    def __assign_unassigned_items(self,
                                  action: str,
                                  collections: List[str],
                                  dataset_id: str = None,
                                  item_id: str = None,
                                  filters: entities.Filters = None
                                  ) -> bool:
        if not isinstance(collections, list):
            raise ValueError("collections must be a list.")
        if dataset_id is None and self._dataset is not None:
            dataset_id = self.dataset.id
        if item_id is None and self._item is not None:
            item_id = self.item.id
            # Build the query structure
        query = {}
        if item_id is not None:
            query = entities.Filters(field="id", values=[item_id], operator=entities.FiltersOperations.IN,
                                     method=entities.FiltersMethod.AND).prepare().get("filter")
        elif filters is not None:
            query = filters.prepare().get("filter")

        # Create the payload
        payload = {
            "query": {"filter": query},
            "collections": collections,
        }

        # Make the API request to assign the item
        success, response = self._client_api.gen_request(
            req_type="post",
            path=f"/datasets/{dataset_id}/items/collections/{action}",
            json_req=payload,
        )

        if success:
            # Wait for the operation to complete
            command = entities.Command.from_json(_json=response.json(), client_api=self._client_api)
            command.wait()
            return True
        else:
            raise exceptions.PlatformException(f"Failed to assign item to collections: {response}")

    def assign(
            self,
            collections: List[str],
            dataset_id: str = None,
            item_id: str = None,
            filters: entities.Filters = None
    ) -> bool:
        """
        Assign an item to a collection. Creates the collection if it does not exist.

        :param collections: List of the collections to assign the item to.
        :param dataset_id: ID of the dataset.
        :param item_id: (Optional) ID of the item to assign. If not provided, all items in the dataset will be updated.
        :param filters: (Optional) Filters of items to assign to the collections.
        :return: True if the assignment was successful, otherwise raises an exception.
        """
        return self.__assign_unassigned_items(
            action="bulk-add",
            collections=collections,
            dataset_id=dataset_id,
            item_id=item_id,
            filters=filters
        )

    def unassign(self,
                 collections: List[str],
                 dataset_id: str = None,
                 item_id: str = None,
                 filters: entities.Filters = None) -> bool:
        """
        Unassign an item from a collection.
        :param collections: List of collection names to unassign.
        :param dataset_id: ID of the dataset.
        :param item_id: ID of the item.
        :param filters: (Optional) Filters of items to unassign from the collections.
        """
        return self.__assign_unassigned_items(
            action="bulk-remove",
            collections=collections,
            dataset_id=dataset_id,
            item_id=item_id,
            filters=filters
        )

    def _single_collection(self, data: dict, name: str):
        """
        Retrieves the key-value pair from the dictionary where the collection's name matches the given name.

        :param data: A dictionary containing collection data in the format:
                    { "metadata.system.collections.c0": {"name": "Justice League"}, ... }
        :param name: The name of the collection to find.
        :return: The key-value pair where the name matches, or None if not found.
        """
        for key, value in data.items():
            if value.get("name") == name:
                return {key: value}
        return None

    def _list_collections(self, data: dict):
        """
        Create a list of Collection entities from the dataset JSON.

        :param data: The flat JSON containing collection data in the format:
                    { "metadata.system.collections.c0": {"name": "Justice League"}, ... }
        :return: A list of Collection entities.
        """
        collections = []
        for full_key, value in data.items():
            if "metadata.system.collections" in full_key:
                # Strip the prefix
                key = full_key.replace("metadata.system.collections.", "")
                collection_name = value.get("name")
                collections.append({"key": key, "name": collection_name})
        return collections

    def get_name_by_key(self, key: str) -> str:
        """
        Get the name of a collection by its key.

        :param key: The key of the collection (e.g., 'c0', 'c1').
        :return: The name of the collection if it exists; otherwise, an empty string.
        """
        # Assuming collections is a list of dictionaries
        collections = self.list_all_collections()
        for collection in collections:
            if collection.get("key") == key:
                return collection.get("name", "")
        return ""



from collections import namedtuple
from enum import Enum
import traceback
import logging
import attr
from .. import repositories, entities
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class DatasetSubsetType(str, Enum):
    """Available types for dataset subsets"""
    TRAIN = 'train'
    VALIDATION = 'validation'
    TEST = 'test'


class ModelStatus(str, Enum):
    """Available types for model status"""
    CREATED = "created",
    PRE_TRAINED = "pre-trained",
    PENDING = "pending",
    TRAINING = "training",
    TRAINED = "trained",
    DEPLOYED = "deployed",
    FAILED = "failed",
    CLONING = "cloning"


class PlotSample:
    def __init__(self, figure, legend, x, y):
        """
        Create a single metric sample for Model

        :param figure: figure name identifier
        :param legend: line name identifier
        :param x: x value for the current sample
        :param y: y value for the current sample
        """
        self.figure = figure
        self.legend = legend
        self.x = x
        self.y = y

    def to_json(self) -> dict:
        _json = {'figure': self.figure,
                 'legend': self.legend,
                 'data': {'x': self.x,
                          'y': self.y}}
        return _json


# class MatrixSample:
#     def __init__(self, figure, legend, x, y):
#         """
#         Create a single metric sample for Model
#
#         :param figure: figure name identifier
#         :param legend: line name identifier
#         :param x: x value for the current sample
#         :param y: y value for the current sample
#         """
#         self.figure = figure
#         self.legend = legend
#         self.x = x
#         self.y = y
#
#     def to_json(self) -> dict:
#         _json = {'figure': self.figure,
#                  'legend': self.legend,
#                  'data': {'x': self.x,
#                           'y': self.y}}
#         return _json


@attr.s
class Model(entities.BaseEntity):
    """
    Model object
    """
    # platform
    id = attr.ib()
    creator = attr.ib()
    created_at = attr.ib()
    updated_at = attr.ib(repr=False)
    model_artifacts = attr.ib()
    name = attr.ib()
    description = attr.ib()
    ontology_id = attr.ib(repr=False)
    labels = attr.ib()
    status = attr.ib()
    tags = attr.ib()
    configuration = attr.ib()
    metadata = attr.ib()
    input_type = attr.ib()
    output_type = attr.ib()
    module_name = attr.ib()

    url = attr.ib()
    scope = attr.ib()
    version = attr.ib()
    context = attr.ib()
    status_logs = attr.ib()

    # name change
    package_id = attr.ib(repr=False)
    project_id = attr.ib()
    dataset_id = attr.ib(repr=False)

    # sdk
    _project = attr.ib(repr=False)
    _package = attr.ib(repr=False)
    _dataset = attr.ib(repr=False)
    _feature_set = attr.ib(repr=False)
    _client_api = attr.ib(type=ApiClient, repr=False)
    _repositories = attr.ib(repr=False)
    _ontology = attr.ib(repr=False, default=None)
    updated_by = attr.ib(default=None)
    app = attr.ib(default=None)

    @staticmethod
    def _protected_from_json(_json, client_api, project=None, package=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform representation of Model
        :param client_api: ApiClient entity
        :param project: project that owns the model
        :param package: package entity of the model
        :param is_fetched: is Entity fetched from Platform
        :return: Model entity
        """
        try:
            model = Model.from_json(_json=_json,
                                    client_api=client_api,
                                    project=project,
                                    package=package,
                                    is_fetched=is_fetched)
            status = True
        except Exception:
            model = traceback.format_exc()
            status = False
        return status, model

    @classmethod
    def from_json(cls, _json, client_api, project=None, package=None, is_fetched=True):
        """
        Turn platform representation of model into a model entity

        :param _json: platform representation of model
        :param client_api: ApiClient entity
        :param project: project that owns the model
        :param package: package entity of the model
        :param is_fetched: is Entity fetched from Platform
        :return: Model entity
        """
        if project is not None:
            if project.id != _json.get('context', {}).get('project', None):
                logger.warning("Model's project is different then the input project")
                project = None

        if package is not None:
            if package.id != _json.get('packageId', None):
                logger.warning("Model's package is different then the input package")
                package = None

        model_artifacts = [entities.Artifact.from_json(_json=artifact,
                                                       client_api=client_api,
                                                       project=project)
                           for artifact in _json.get('artifacts', list())]

        inst = cls(
            configuration=_json.get('configuration', None),
            description=_json.get('description', None),
            status=_json.get('status', None),
            tags=_json.get('tags', None),
            metadata=_json.get('metadata', dict()),
            project_id=_json.get('context', {}).get('project', None),
            dataset_id=_json.get('datasetId', None),
            package_id=_json.get('packageId', None),
            model_artifacts=model_artifacts,
            labels=_json.get('labels', None),
            ontology_id=_json.get('ontology_id', None),
            created_at=_json.get('createdAt', None),
            updated_at=_json.get('updatedAt', None),
            creator=_json.get('context', {}).get('creator', None),
            client_api=client_api,
            name=_json.get('name', None),
            project=project,
            package=package,
            dataset=None,
            feature_set=None,
            id=_json.get('id', None),
            url=_json.get('url', None),
            scope=_json.get('scope', entities.EntityScopeLevel.PROJECT),
            version=_json.get('version', '1.0.0'),
            context=_json.get('context', {}),
            input_type=_json.get('inputType', None),
            output_type=_json.get('outputType', None),
            module_name=_json.get('moduleName', None),
            updated_by=_json.get('updatedBy', None),
            app=_json.get('app', None),
            status_logs=_json.get('statusLogs', []),
        )
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Get the dict of Model

        :return: platform json of model
        :rtype: dict
        """
        _json = attr.asdict(self,
                            filter=attr.filters.exclude(attr.fields(Model)._project,
                                                        attr.fields(Model)._package,
                                                        attr.fields(Model)._dataset,
                                                        attr.fields(Model)._ontology,
                                                        attr.fields(Model)._repositories,
                                                        attr.fields(Model)._feature_set,
                                                        attr.fields(Model)._client_api,
                                                        attr.fields(Model).package_id,
                                                        attr.fields(Model).project_id,
                                                        attr.fields(Model).dataset_id,
                                                        attr.fields(Model).ontology_id,
                                                        attr.fields(Model).model_artifacts,
                                                        attr.fields(Model).created_at,
                                                        attr.fields(Model).updated_at,
                                                        attr.fields(Model).input_type,
                                                        attr.fields(Model).output_type,
                                                        attr.fields(Model).updated_by,
                                                        attr.fields(Model).app,
                                                        attr.fields(Model).status_logs
                                                        ))
        _json['packageId'] = self.package_id
        _json['datasetId'] = self.dataset_id
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at
        _json['inputType'] = self.input_type
        _json['outputType'] = self.output_type
        _json['moduleName'] = self.module_name

        model_artifacts = list()
        for artifact in self.model_artifacts:
            if artifact.type in ['file', 'dir']:
                artifact = {'type': 'item',
                            'itemId': artifact.id}
            else:
                artifact = artifact.to_json(as_artifact=True)
            model_artifacts.append(artifact)
        _json['artifacts'] = model_artifacts

        if self.updated_by:
            _json['updatedBy'] = self.updated_by
        if self.app:
            _json['app'] = self.app
        if self.status_logs:
            _json['statusLogs'] = self.status_logs

        return _json

    ############
    # entities #
    ############
    @property
    def project(self):
        if self._project is None:
            self._project = self.projects.get(project_id=self.project_id, fetch=None)
            self._repositories = self.set_repositories()  # update the repos with the new fetched entity
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def feature_set(self) -> 'entities.FeatureSet':
        if self._feature_set is None:
            filters = entities.Filters(field='modelId',
                                       values=self.id,
                                       resource=entities.FiltersResource.FEATURE_SET)
            feature_sets = self.project.feature_sets.list(filters=filters)
            if feature_sets.items_count > 1:
                logger.warning("Found more than one feature set associated with model entity. Returning first result."
                               "Set feature_set if other feature set entity is needed.")
                self._feature_set = feature_sets.items[0]
            elif feature_sets.items_count == 1:
                self._feature_set = feature_sets.items[0]
            else:
                self._feature_set = None
        return self._feature_set

    @feature_set.setter
    def feature_set(self, feature_set: 'entities.FeatureSet'):
        if not isinstance(feature_set, entities.FeatureSet):
            raise ValueError("feature_set must be of type dl.FeatureSet")
        else:
            self._feature_set = feature_set

    @property
    def package(self):
        if self._package is None:
            try:
                if self.app:
                    self._package = self.dpks.get_revisions(dpk_id=self.app['dpkId'], version=self.app['dpkVersion'])
                else:
                    self._package = self.packages.get(package_id=self.package_id)
            except Exception as e:
                error = e
                try:
                    self._package = self.dpks.get(dpk_id=self.package_id)
                except Exception:
                    raise error
            self._repositories = self.set_repositories()  # update the repos with the new fetched entity
        assert isinstance(self._package, (entities.Package, entities.Dpk))
        return self._package

    @property
    def dataset(self):
        if self._dataset is None:
            if self.dataset_id is None:
                raise RuntimeError("Model {!r} has no dataset. Can be used only for inference".format(self.id))
            self._dataset = self.datasets.get(dataset_id=self.dataset_id, fetch=None)
            self._repositories = self.set_repositories()  # update the repos with the new fetched entity
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @property
    def ontology(self):
        if self._ontology is None:
            if self.ontology_id is None:
                raise RuntimeError("Model {!r} has no ontology.".format(self.id))
            self._ontology = self.ontologies.get(ontology_id=self.ontology_id)
        assert isinstance(self._ontology, entities.Ontology)
        return self._ontology

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['projects', 'datasets', 'models', 'packages', 'ontologies', 'artifacts',
                                       'metrics', 'dpks', 'services'])

        r = reps(projects=repositories.Projects(client_api=self._client_api),
                 datasets=repositories.Datasets(client_api=self._client_api,
                                                project=self._project),
                 models=repositories.Models(client_api=self._client_api,
                                            project=self._project,
                                            project_id=self.project_id,
                                            package=self._package),
                 packages=repositories.Packages(client_api=self._client_api,
                                                project=self._project),
                 ontologies=repositories.Ontologies(client_api=self._client_api,
                                                    project=self._project,
                                                    dataset=self._dataset),
                 artifacts=repositories.Artifacts(client_api=self._client_api,
                                                  project=self._project,
                                                  project_id=self.project_id,
                                                  model=self),
                 metrics=repositories.Metrics(client_api=self._client_api,
                                              model=self),
                 dpks=repositories.Dpks(client_api=self._client_api),
                 services=repositories.Services(client_api=self._client_api,
                                                project=self._project,
                                                project_id=self.project_id,
                                                model_id=self.id,
                                                model=self),
                 )
        return r

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/model/{}".format(self.project_id, self.id))

    @property
    def projects(self):
        assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def datasets(self):
        assert isinstance(self._repositories.datasets, repositories.Datasets)
        return self._repositories.datasets

    @property
    def models(self):
        assert isinstance(self._repositories.models, repositories.Models)
        return self._repositories.models

    @property
    def packages(self):
        assert isinstance(self._repositories.packages, repositories.Packages)
        return self._repositories.packages

    @property
    def dpks(self):
        assert isinstance(self._repositories.dpks, repositories.Dpks)
        return self._repositories.dpks

    @property
    def ontologies(self):
        assert isinstance(self._repositories.ontologies, repositories.Ontologies)
        return self._repositories.ontologies

    @property
    def artifacts(self):
        assert isinstance(self._repositories.artifacts, repositories.Artifacts)
        return self._repositories.artifacts

    @property
    def metrics(self):
        assert isinstance(self._repositories.metrics, repositories.Metrics)
        return self._repositories.metrics

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    @property
    def id_to_label_map(self):
        # default
        if 'id_to_label_map' not in self.configuration:
            if not (self.dataset_id == 'null' or self.dataset_id is None):
                self.labels = [flat_key for flat_key, _ in self.dataset.labels_flat_dict.items()]
            self.configuration['id_to_label_map'] = {int(idx): lbl for idx, lbl in enumerate(self.labels)}
        # use existing
        else:
            self.configuration['id_to_label_map'] = {int(idx): lbl for idx, lbl in
                                                     self.configuration['id_to_label_map'].items()}
        return self.configuration['id_to_label_map']

    @id_to_label_map.setter
    def id_to_label_map(self, mapping: dict):
        self.configuration['id_to_label_map'] = {int(idx): lbl for idx, lbl in mapping.items()}

    @property
    def label_to_id_map(self):
        if 'label_to_id_map' not in self.configuration:
            self.configuration['label_to_id_map'] = {v: int(k) for k, v in self.id_to_label_map.items()}
        return self.configuration['label_to_id_map']

    @label_to_id_map.setter
    def label_to_id_map(self, mapping: dict):
        self.configuration['label_to_id_map'] = {v: int(k) for k, v in mapping.items()}

    ###########
    # methods #
    ###########

    def add_subset(
        self, subset_name: str, subset_filter=None, subset_annotation_filter=None
    ):
        """
        Adds a subset for the model, specifying a subset of the model's dataset that could be used for training or
        validation. Optionally also adds an annotations subset.

        :param str subset_name: the name of the subset
        :param subset_filter: filtering for items subset. Can be `entities.Filters`, `dict`, or `None`
        :param subset_annotation_filter: optional filtering for annotations subset. Can be `entities.Filters`, `dict`, or `None`

        """
        self.models.add_subset(self, subset_name, subset_filter, subset_annotation_filter)

    def delete_subset(self, subset_name: str):
        """
        Removes a subset from the model's metadata (both subsets and annotationsSubsets).

        :param str subset_name: the name of the subset

        """
        self.models.delete_subset(self, subset_name)

    def update(self, system_metadata=False, reload_services=True):
        """
        Update Models changes to platform

        :param bool system_metadata: bool - True, if you want to change metadata system
        :param bool reload_services: bool - True, if you want to update the services with the new model
        :return: Models entity
        """
        return self.models.update(model=self,
                                  system_metadata=system_metadata,
                                  reload_services=reload_services
                                  )

    def open_in_web(self):
        """
        Open the model in web platform

        :return:
        """
        self._client_api._open_in_web(url=self.platform_url)

    def delete(self):
        """
        Delete Model object

        :return: True
        """
        return self.models.delete(model=self)

    def clone(self,
              model_name: str,
              dataset: entities.Dataset = None,
              configuration: dict = None,
              status=None,
              scope=None,
              project_id: str = None,
              labels: list = None,
              description: str = None,
              tags: list = None,
              train_filter: entities.Filters = None,
              validation_filter: entities.Filters = None,
              annotations_train_filter: entities.Filters = None,
              annotations_validation_filter: entities.Filters = None,
              wait=True
              ):
        """
        Clones and creates a new model out of existing one

        :param str model_name: `str` new model name
        :param str dataset: dataset object for the cloned model
        :param dict configuration: `dict` (optional) if passed replaces the current configuration
        :param str status: `str` (optional) set the new status
        :param str scope: `str` (optional) set the new scope. default is "project"
        :param str project_id: `str` specify the project id to create the new model on (if other than the source model)
        :param list labels:  `list` of `str` - label of the model
        :param str description: `str` description of the new model
        :param list tags:  `list` of `str` - label of the model
        :param dtlpy.entities.filters.Filters train_filter: Filters entity or a dictionary to define the items' scope in the specified dataset_id for the model train
        :param dtlpy.entities.filters.Filters validation_filter: Filters entity or a dictionary to define the items' scope in the specified dataset_id for the model validation
        :param dtlpy.entities.filters.Filters annotations_train_filter: Filters entity or a dictionary to define the annotations' scope in the specified dataset_id for the model train
        :param dtlpy.entities.filters.Filters annotations_validation_filter: Filters entity or a dictionary to define the annotations' scope in the specified dataset_id for the model validation
        :param bool wait: `bool` wait for the model to be ready before returning

        :return: dl.Model which is a clone version of the existing model
        """
        return self.models.clone(from_model=self,
                                 model_name=model_name,
                                 project_id=project_id,
                                 dataset=dataset,
                                 scope=scope,
                                 status=status,
                                 configuration=configuration,
                                 labels=labels,
                                 description=description,
                                 tags=tags,
                                 train_filter=train_filter,
                                 validation_filter=validation_filter,
                                 annotations_train_filter=annotations_train_filter,
                                 annotations_validation_filter=annotations_validation_filter,
                                 wait=wait
                                 )

    def train(self, service_config=None):
        """
        Train the model in the cloud. This will create a service and will run the adapter's train function as an execution

        :param dict service_config : Service object as dict. Contains the spec of the default service to create.
        :return:
        """
        return self.models.train(model_id=self.id, service_config=service_config)

    def evaluate(self, dataset_id, filters: entities.Filters = None, service_config=None):
        """
        Evaluate Model, provide data to evaluate the model on You can also provide specific config for the deployed service

        :param dict service_config : Service object as dict. Contains the spec of the default service to create.
        :param str dataset_id: ID of the dataset to evaluate
        :param entities.Filters filters: dl.Filter entity to run the predictions on
        :return:
        """
        return self.models.evaluate(model_id=self.id,
                                    dataset_id=dataset_id,
                                    filters=filters,
                                    service_config=service_config)

    def predict(self, item_ids=None, dataset_id=None):
        """
        Run model prediction with items

        :param item_ids: a list of item id to run the prediction.
        :param dataset_id: dataset id to run the prediction on
        :return:
        """
        return self.models.predict(model=self, item_ids=item_ids, dataset_id=dataset_id)

    def embed(self, item_ids):
        """
        Run model embed with items

        :param item_ids: a list of item id to run the embed.
        :return:
        """
        return self.models.embed(model=self, item_ids=item_ids)

    def embed_datasets(self, dataset_ids, attach_trigger=False):
        """
        Run model embed with datasets

        :param dataset_ids: a list of dataset id to run the embed.
        :param attach_trigger: bool - True, if you want to activate the trigger
        :return:
        """
        return self.models.embed_datasets(model=self, dataset_ids=dataset_ids, attach_trigger=attach_trigger)

    def deploy(self, service_config=None) -> entities.Service:
        """
        Deploy a trained model. This will create a service that will execute predictions

        :param dict service_config : Service object as dict. Contains the spec of the default service to create.

        :return: dl.Service: The deployed service
        """
        return self.models.deploy(model_id=self.id, service_config=service_config)

    def wait_for_model_ready(self):
        """
        Wait for model to be ready

        :return:
        """
        return self.models.wait_for_model_ready(model=self)

    def log(self,
            service=None,
            size=None,
            checkpoint=None,
            start=None,
            end=None,
            follow=False,
            text=None,
            execution_id=None,
            function_name=None,
            replica_id=None,
            system=False,
            view=True,
            until_completed=True,
            model_operation: str = None,
            ):
        """
        Get service logs

        :param service: service object
        :param int size: size
        :param dict checkpoint: the information from the lst point checked in the service
        :param str start: iso format time
        :param str end: iso format time
        :param bool follow: if true, keep stream future logs
        :param str text: text
        :param str execution_id: execution id
        :param str function_name: function name
        :param str replica_id: replica id
        :param bool system: system
        :param bool view: if true, print out all the logs
        :param bool until_completed: wait until completed
        :param str model_operation: model operation action
        :return: ServiceLog entity
        :rtype: ServiceLog

        **Example**:

        .. code-block:: python

            service_log = service.log()
        """
        return self.services.log(service=service,
                                 size=size,
                                 checkpoint=checkpoint,
                                 start=start,
                                 end=end,
                                 follow=follow,
                                 execution_id=execution_id,
                                 function_name=function_name,
                                 replica_id=replica_id,
                                 system=system,
                                 text=text,
                                 view=view,
                                 until_completed=until_completed,
                                 model_id=self.id,
                                 model_operation=model_operation,
                                 project_id=self.project_id)



from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from PIL import Image
import collections.abc
import numpy as np
import collections
import logging
import shutil
import json
import copy
import tqdm
import sys
import os
import re
from ... import entities

logger = logging.getLogger(name='dtlpy')


class DataItem(dict):
    def __init__(self, *args, **kwargs):
        super(DataItem, self).__init__(*args, **kwargs)

    @property
    def image_filepath(self):
        return self['image_filepath']

    @image_filepath.setter
    def image_filepath(self, val):
        self['image_filepath'] = val


class DatasetGenerator:

    def __init__(self,
                 dataset_entity: entities.Dataset,
                 annotation_type: entities.AnnotationType,
                 item_type: list = None,
                 filters: entities.Filters = None,
                 data_path=None,
                 overwrite=False,
                 id_to_label_map=None,
                 label_to_id_map=None,
                 transforms=None,
                 transforms_callback=None,
                 num_workers=0,
                 batch_size=None,
                 collate_fn=None,
                 shuffle=True,
                 seed=None,
                 to_categorical=False,
                 to_mask=False,
                 class_balancing=False,
                 # debug flags
                 return_originals=False,
                 ignore_empty=True
                 ) -> None:
        """
        Base Dataset Generator to build and iterate over images and annotations

        * Mapping Labels *
        To set a label mapping from labels to id you can use the `label_to_id_map` or `id_to_label_map`.
        NOTE: if they are not i.i.d you'll need to input both.
        In semantic, a `$default` label should be added so that the background (and all unlabeled pixels) will be
        mapped to the model's inputs

        label_to_id_map = {'cat': 1,
                           'dog': 1,
                           '$default': 0}
        id_to_label_map = {1: 'cats_and_dogs',
                           0: 'background'}

        :param dataset_entity: dl.Dataset entity
        :param annotation_type: dl.AnnotationType - type of annotation to load from the annotated dataset
        :param item_type: list of file extension to load. default: ['jpg', 'jpeg', 'png', 'bmp']
        :param filters: dl.Filters - filtering entity to filter the dataset items
        :param data_path: Path to Dataloop annotations (root to "item" and "json").
        :param overwrite:
        :param dict id_to_label_map: Optional, {id: label_string} dictionary, default taken from dataset
        :param dict label_to_id_map: Optional, {label_string: id} dictionary
        :param transforms: Optional transform to be applied on a sample. list, imgaug.Sequence or torchvision.transforms.Compose
        :param transforms_callback: Optional function to handle the callback of each batch.
        look at default_transforms_callback for more information. available: imgaug_transforms_callback, torchvision_transforms_callback
        :param num_workers: Optional - number of separate threads to load the images
        :param batch_size: (int, optional): how many samples per batch to load, if not none - items will always be a list
        :param collate_fn: Optional - merges a list of samples to form a mini-batch of Tensor(s).
        :param shuffle: Whether to shuffle the data (default: True) If set to False, sorts the data in alphanumeric order.
        :param seed: Optional random seed for shuffling and transformations.
        :param to_categorical: convert label id to categorical format
        :param to_mask: convert annotations to an instance mask (will be true for SEGMENTATION)
        :param class_balancing: if True - performing random over-sample with class ids as the target to balance training data
        :param return_originals: bool - If True, return ALSO images and annotations before transformations (for debug)
        :param ignore_empty: bool - If True, generator will NOT collect items without annotations
        """
        self._dataset_entity = dataset_entity

        # default item types (extension for now)
        if item_type is None:
            item_type = ['jpg', 'jpeg', 'png', 'bmp']
        if not isinstance(item_type, list):
            item_type = [item_type]
        self.item_type = item_type

        # id labels mapping
        if label_to_id_map is None and id_to_label_map is None:
            # if both are None - take from dataset
            label_to_id_map = dataset_entity.instance_map
            id_to_label_map = {int(v): k for k, v in label_to_id_map.items()}
        else:
            # one or both is NOT None
            if label_to_id_map is None:
                # set label_to_id_map from the other
                label_to_id_map = {v: int(k) for k, v in id_to_label_map.items()}
            if id_to_label_map is None:
                # set id_to_label_map from the other
                id_to_label_map = {int(v): k for k, v in label_to_id_map.items()}
            # put it on the local ontology for the annotations download
            dataset_entity._get_ontology().instance_map = label_to_id_map
        self.id_to_label_map = id_to_label_map
        self.label_to_id_map = label_to_id_map

        # if annotation type is segmentation - to_mask must be True
        if annotation_type == entities.AnnotationType.SEGMENTATION:
            to_mask = True

        if data_path is None:
            data_path = os.path.join(os.path.expanduser('~'),
                                     '.dataloop',
                                     'datasets',
                                     "{}_{}".format(dataset_entity.name,
                                                    dataset_entity.id))
        download = False
        if os.path.isdir(data_path):
            if overwrite:
                logger.warning('overwrite flag is True! deleting and overwriting')
                shutil.rmtree(data_path)
                download = True
        else:
            download = True
        if download:
            annotation_options = [entities.ViewAnnotationOptions.JSON]
            if to_mask is True:
                annotation_options.append(entities.ViewAnnotationOptions.INSTANCE)
            _ = dataset_entity.items.download(filters=filters,
                                              local_path=data_path,
                                              thickness=-1,
                                              annotation_options=annotation_options)
        self.root_dir = data_path
        self._items_path = Path(self.root_dir).joinpath('items')
        self._json_path = Path(self.root_dir).joinpath('json')
        self._mask_path = Path(self.root_dir).joinpath('instance')
        self._transforms = transforms
        self._transforms_callback = transforms_callback
        if self._transforms is not None and self._transforms_callback is None:
            # use default callback
            self._transforms_callback = default_transforms_callback

        self.annotation_type = annotation_type
        self.num_workers = num_workers
        self.to_categorical = to_categorical
        self.num_classes = len(label_to_id_map)
        self.shuffle = shuffle
        self.seed = seed
        self.to_mask = to_mask
        self.batch_size = batch_size
        self.collate_fn = collate_fn
        self.class_balancing = class_balancing
        # inits
        self.data_items = list()
        # flags
        self.return_originals = return_originals
        self.ignore_empty = ignore_empty

        ####################
        # Load annotations #
        ####################
        self.load_annotations()

    @property
    def dataset_entity(self):
        assert isinstance(self._dataset_entity, entities.Dataset)
        return self._dataset_entity

    @dataset_entity.setter
    def dataset_entity(self, val):
        assert isinstance(val, entities.Dataset)
        self._dataset_entity = val

    @property
    def n_items(self):
        return len(self.data_items)

    def _load_single(self, image_filepath, pbar=None):
        try:
            is_empty = False
            item_info = DataItem()
            # add image path
            item_info.image_filepath = str(image_filepath)
            if os.stat(image_filepath).st_size < 5:
                logger.warning('IGNORING corrupted image: {!r}'.format(image_filepath))
                return None, True
            # get "platform" path
            rel_path = image_filepath.relative_to(self._items_path)
            # replace suffix to JSON
            rel_path_wo_png_ext = rel_path.with_suffix('.json')
            # create local path
            annotation_filepath = Path(self._json_path, rel_path_wo_png_ext)

            if os.path.isfile(annotation_filepath):
                with open(annotation_filepath, 'r') as f:
                    data = json.load(f)
                    if 'id' in data:
                        item_id = data.get('id')
                    elif '_id' in data:
                        item_id = data.get('_id')
                    annotations = entities.AnnotationCollection.from_json(data)
            else:
                item_id = ''
                annotations = None
            item_info.update(item_id=item_id)
            if self.annotation_type is not None:
                # add item id from json
                polygon_coordinates = list()
                box_coordinates = list()
                classes_ids = list()
                labels = list()
                if annotations is not None:
                    for annotation in annotations:
                        if 'user' in annotation.metadata and \
                                'model' in annotation.metadata['user']:
                            # and 'name' in annotation.metadata['user']['model']:
                            # Do not use prediction annotations in the data generator
                            continue
                        if annotation.type == self.annotation_type:
                            if annotation.label not in self.label_to_id_map:
                                logger.warning(
                                    'Missing label {!r} in label_to_id_map. Skipping.. Use label_to_id_map for other behaviour'.format(
                                        annotation.label))
                            else:
                                classes_ids.append(self.label_to_id_map[annotation.label])
                            labels.append(annotation.label)
                            box_coordinates.append(np.asarray([annotation.left,
                                                               annotation.top,
                                                               annotation.right,
                                                               annotation.bottom]))
                            if self.annotation_type == entities.AnnotationType.POLYGON:
                                polygon_coordinates.append(annotation.geo)
                            if annotation.type not in [entities.AnnotationType.CLASSIFICATION,
                                                       entities.AnnotationType.SEGMENTATION,
                                                       entities.AnnotationType.BOX,
                                                       entities.AnnotationType.POLYGON]:
                                raise ValueError('unsupported annotation type: {}'.format(annotation.type))
                dtype = object if self.annotation_type == entities.AnnotationType.POLYGON else None
                # reorder for output
                item_info.update({entities.AnnotationType.BOX.value: np.asarray(box_coordinates).astype(float),
                                  entities.AnnotationType.CLASSIFICATION.value: np.asarray(classes_ids),
                                  entities.AnnotationType.POLYGON.value: np.asarray(polygon_coordinates, dtype=dtype),
                                  'labels': labels})
                if len(item_info[entities.AnnotationType.CLASSIFICATION.value]) == 0:
                    logger.debug('Empty annotation (nothing matched label_to_id_map) for image filename: {}'.format(
                        image_filepath))
                    is_empty = True
            if self.to_mask:
                # get "platform" path
                rel_path = image_filepath.relative_to(self._items_path)
                # replace suffix to PNG
                rel_path_wo_png_ext = rel_path.with_suffix('.png')
                # create local path
                mask_filepath = Path(self._mask_path, rel_path_wo_png_ext)
                if not os.path.isfile(mask_filepath):
                    logger.debug('Empty annotation for image filename: {}'.format(image_filepath))
                    is_empty = True
                item_info.update({entities.AnnotationType.SEGMENTATION.value: str(mask_filepath)})
            item_info.update(annotation_filepath=str(annotation_filepath))
            return item_info, is_empty
        except Exception:
            logger.exception('failed loading item in generator! {!r}'.format(image_filepath))
            return None, True
        finally:
            if pbar is not None:
                pbar.update()

    def load_annotations(self):
        logger.info(f"Collecting items with the following extensions: {self.item_type}")
        files = list()
        for ext in self.item_type:
            # build regex to ignore extension case
            regex = '*.{}'.format(''.join(['[{}{}]'.format(letter.lower(), letter.upper()) for letter in ext]))
            files.extend(self._items_path.rglob(regex))

        pool = ThreadPoolExecutor(max_workers=32)
        jobs = list()
        pbar = tqdm.tqdm(total=len(files),
                         desc='Loading Data Generator',
                         disable=self.dataset_entity._client_api.verbose.disable_progress_bar,
                         file=sys.stdout)
        for image_filepath in files:
            jobs.append(pool.submit(self._load_single,
                                    image_filepath=image_filepath,
                                    pbar=pbar))
        outputs = [job.result() for job in jobs]
        pbar.close()

        n_items = len(outputs)
        n_empty_items = sum([1 for _, is_empty in outputs if is_empty is True])

        output_msg = 'Done loading items. Total items loaded: {}.'.format(n_items)
        if n_empty_items > 0:
            output_msg += '{action} {n_empty_items} items without annotations'.format(
                action='IGNORING' if self.ignore_empty else 'INCLUDING',
                n_empty_items=n_empty_items)

        if self.ignore_empty:
            # take ONLY non-empty
            data_items = [data_item for data_item, is_empty in outputs if is_empty is False]
        else:
            # take all
            data_items = [data_item for data_item, is_empty in outputs]

        self.data_items = data_items
        if len(self.data_items) == 0:
            logger.warning(output_msg)
        else:
            logger.info(output_msg)
        ###################
        # class balancing #
        ###################
        labels = [label for item in self.data_items for label in item.get('labels', list())]
        logger.info(f"Data Generator labels balance statistics: {collections.Counter(labels)}")
        if self.class_balancing:
            try:
                from imblearn.over_sampling import RandomOverSampler
            except Exception:
                logger.error(
                    'Class balancing is ON but missing "imbalanced-learn". run "pip install -U imbalanced-learn" and try again')
                raise
            logger.info('Class balance is on!')
            class_ids = [class_id for item in self.data_items for class_id in item['class']]
            dummy_inds = [i_item for i_item, item in enumerate(self.data_items) for _ in item['class']]
            over_sampler = RandomOverSampler(random_state=42)
            X_res, y_res = over_sampler.fit_resample(np.asarray(dummy_inds).reshape(-1, 1), np.asarray(class_ids))
            over_sampled_data_items = [self.data_items[i] for i in X_res.flatten()]
            oversampled_labels = [label for item in over_sampled_data_items for label in item['labels']]
            logger.info(f"Data Generator labels after oversampling: {collections.Counter(oversampled_labels)}")
            self.data_items = over_sampled_data_items

        if self.shuffle:
            if self.seed is None:
                self.seed = 256
            np.random.seed(self.seed)
            np.random.shuffle(self.data_items)

    def transform(self, image, target=None):
        if self._transforms is not None:
            image, target = self._transforms_callback(transforms=self._transforms,
                                                      image=image,
                                                      target=target,
                                                      annotation_type=self.annotation_type)
        return image, target

    def _to_dtlpy(self, targets, labels=None):
        annotations = entities.AnnotationCollection(item=None)
        annotations._dataset = self._dataset_entity
        if labels is None:
            labels = [None] * len(targets)
        if self.to_mask is True:
            for label, label_ind in self.label_to_id_map.items():
                target = targets == label_ind
                if np.any(target):
                    annotations.add(annotation_definition=entities.Segmentation(geo=target,
                                                                                label=label))
        elif self.annotation_type == entities.AnnotationType.BOX:
            for target, label in zip(targets, labels):
                annotations.add(annotation_definition=entities.Box(left=target[0],
                                                                   top=target[1],
                                                                   right=target[2],
                                                                   bottom=target[3],
                                                                   label=label))
        elif self.annotation_type == entities.AnnotationType.CLASSIFICATION:
            for target, label in zip(targets, labels):
                annotations.add(annotation_definition=entities.Classification(label=label))
        elif self.annotation_type == entities.AnnotationType.POLYGON:
            for target, label in zip(targets, labels):
                annotations.add(annotation_definition=entities.Polygon(label=label,
                                                                       geo=target.astype(float)))
        else:
            raise ValueError('unsupported annotation type: {}'.format(self.annotation_type))
        # set dataset for color
        for annotation in annotations:
            annotation._dataset = self._dataset_entity
        return annotations

    def visualize(self, idx=None, return_output=False, plot=True):
        if not self.__len__():
            raise ValueError('no items selected, cannot preform visualization')
        import matplotlib.pyplot as plt
        if idx is None:
            idx = np.random.randint(self.__len__())
        if self.batch_size is not None:
            raise ValueError('can visualize only of batch_size in None')
        data_item = self.__getitem__(idx)
        image = Image.fromarray(data_item.get('image'))
        labels = data_item.get('labels')
        targets = data_item.get('annotations')
        annotations = self._to_dtlpy(targets=targets, labels=labels)
        mask = Image.fromarray(annotations.show(height=image.size[1],
                                                width=image.size[0],
                                                alpha=0.8))
        image.paste(mask, (0, 0), mask)
        marked_image = np.asarray(image)
        if plot:
            plt.figure()
            plt.imshow(marked_image)
        if return_output:
            return marked_image, annotations

    def __getsingleitem__(self, idx):
        data_item = copy.deepcopy(self.data_items[idx])

        image_filename = data_item.get('image_filepath')
        image = np.asarray(Image.open(image_filename))
        data_item.update({'image': image})

        annotations = data_item.get(self.annotation_type)
        if self.to_mask is True:
            # if segmentation - read from file
            mask_filepath = data_item.get(entities.AnnotationType.SEGMENTATION)
            annotations = np.asarray(Image.open(mask_filepath).convert('L'))
        if self.to_categorical:
            onehot = np.zeros((annotations.size, self.num_classes + 1))
            onehot[np.arange(annotations.size), annotations] = 1
            annotations = onehot
        data_item.update({'annotations': annotations})

        if self.return_originals is True:
            annotations = []
            if self.annotation_type is not None:
                annotations = data_item.get('annotations')
            data_item.update({'orig_image': image.copy(),
                              'orig_annotations': annotations.copy()})

        ###########################
        # perform transformations #
        ###########################
        if self._transforms is not None:
            annotations = data_item.get('annotations')
            image, annotations = self.transform(image, annotations)
            data_item.update({'image': image,
                              'annotations': annotations})
        return data_item

    def __iter__(self):
        """Create a generator that iterate over the Sequence."""
        for item in (self[i] for i in range(len(self))):
            yield item

    def __len__(self):
        factor = self.batch_size
        if factor is None:
            factor = 1
        return int(np.ceil(self.n_items / factor))

    def __getitem__(self, idx):
        """
            Support single index or a slice.
            Uses ThreadPoolExecutor is num_workers != 0
        """
        to_return = None
        if isinstance(idx, int):
            if self.batch_size is None:
                to_return = self.__getsingleitem__(idx)
            else:
                # if batch_size is define, convert idx to batches
                idx = slice(idx * self.batch_size, min((idx + 1) * self.batch_size, len(self.data_items)))

        if isinstance(idx, slice):
            to_return = list()
            idxs = list(range(idx.start, idx.stop,
                              idx.step if idx.step else 1))
            if self.num_workers == 0:
                for dx in idxs:
                    to_return.append(self.__getsingleitem__(dx))
            else:
                with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
                    for sample in executor.map(lambda i: self.__getsingleitem__(i), idxs):
                        to_return.append(sample)

        if to_return is None:
            raise TypeError('unsupported indexing: list indices must be integers or slices, not {}'.format(type(idx)))

        if self.collate_fn is not None:
            to_return = self.collate_fn(to_return)
        return to_return


np_str_obj_array_pattern = re.compile(r'[SaUO]')

default_collate_err_msg_format = (
    "default_collate: batch must contain tensors, numpy arrays, numbers, "
    "dicts or lists; found {}")


def default_transforms_callback(transforms, image, target, annotation_type):
    """
    Recursive call to perform the augmentations in "transforms"

    :param transforms:
    :param image:
    :param target:
    :param annotation_type:
    :return:
    """
    # get the type string without importing any other package
    transforms_type = type(transforms)

    ############
    # Handle compositions and lists of augmentations with a recursive call
    if transforms_type.__module__ == 'torchvision.transforms.transforms' and transforms_type.__name__ == 'Compose':
        # torchvision compose - convert to list
        image, target = default_transforms_callback(transforms.transforms, image, target, annotation_type)
        return image, target

    if transforms_type.__module__ == 'imgaug.augmenters.meta' and transforms_type.__name__ == 'Sequential':
        # imgaug sequential - convert to list
        image, target = default_transforms_callback(list(transforms), image, target, annotation_type)
        return image, target

    if isinstance(transforms, list):
        for t in transforms:
            image, target = default_transforms_callback(t, image, target, annotation_type)
        return image, target

    ##############
    # Handle single annotations
    if 'imgaug.augmenters' in transforms_type.__module__:
        # handle single imgaug augmentation
        if target is not None and annotation_type is not None:
            # works for batch but running on a single image
            if annotation_type == entities.AnnotationType.BOX:
                image, target = transforms(images=[image], bounding_boxes=[target])
                target = target[0]
            elif annotation_type == entities.AnnotationType.SEGMENTATION:
                # expending to HxWx1 for the imgaug function to work
                target = target[..., None]
                image, target = transforms(images=[image], segmentation_maps=[target])
                target = target[0][:, :, 0]
            elif annotation_type == entities.AnnotationType.POLYGON:
                image, target = transforms(images=[image], polygons=[target])
                target = target[0]
            elif annotation_type == entities.AnnotationType.CLASSIFICATION:
                image = transforms(images=[image])
            else:
                raise ValueError('unsupported annotations type for image augmentations: {}'.format(annotation_type))
            image = image[0]
        else:
            image = transforms(images=[image])
            image = image[0]
    else:
        image = transforms(image)

    return image, target


def collate_default(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""
    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, np.ndarray):
        return np.stack(batch, axis=0)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' and elem_type.__name__ != 'string_':
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(default_collate_err_msg_format.format(elem.dtype))
            return batch
            # return [tf.convert_to_tensor(b) for b in batch]
        elif elem.shape == ():  # scalars
            return batch
    elif isinstance(elem, float):
        return batch
    elif isinstance(elem, int):
        return batch
    elif isinstance(elem, str) or isinstance(elem, bytes) or elem is None:
        return batch
    elif isinstance(elem, collections.abc.Mapping):
        return {key: collate_default([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(collate_default(samples) for samples in zip(*batch)))
    elif isinstance(elem, collections.abc.Sequence):
        transposed = zip(*batch)
        return transposed
    raise TypeError(default_collate_err_msg_format.format(elem_type))


def collate_torch(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""
    import torch
    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        out = None
        if torch.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum(x.numel() for x in batch)
            storage = elem.storage()._new_shared(numel)
            out = elem.new(storage)
        return torch.stack(batch, 0, out=out)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' and elem_type.__name__ != 'string_':
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(default_collate_err_msg_format.format(elem.dtype))
            try:
                return torch.stack([torch.as_tensor(b) for b in batch])
            except RuntimeError:
                return batch
        elif elem.shape == ():  # scalars
            return torch.as_tensor(batch)
    elif isinstance(elem, float):
        return torch.tensor(batch, dtype=torch.float64)
    elif isinstance(elem, int):
        return torch.tensor(batch)
    elif isinstance(elem, str) or isinstance(elem, bytes) or elem is None:
        return batch
    elif isinstance(elem, collections.abc.Mapping):
        return {key: collate_torch([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(collate_torch(samples) for samples in zip(*batch)))
    elif isinstance(elem, collections.abc.Sequence):
        transposed = zip(*batch)
        return transposed

    raise TypeError(default_collate_err_msg_format.format(elem_type))


def collate_tf(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""
    import tensorflow as tf
    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, tf.Tensor):
        return tf.stack(batch, axis=0)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' and elem_type.__name__ != 'string_':
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(default_collate_err_msg_format.format(elem.dtype))
            try:
                return tf.convert_to_tensor(batch)
            except ValueError:
                # failed on orig_image because of a mismatch in the shape (not resizing all the images so cannot stack)
                return batch
                # return [tf.convert_to_tensor(b) for b in batch]
        elif elem.shape == ():  # scalars
            return tf.convert_to_tensor(batch)
    elif isinstance(elem, float):
        return tf.convert_to_tensor(batch, dtype=tf.float64)
    elif isinstance(elem, int):
        return tf.convert_to_tensor(batch)
    elif isinstance(elem, str) or isinstance(elem, bytes) or elem is None:
        return batch
    elif isinstance(elem, collections.abc.Mapping):
        return {key: collate_tf([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(collate_tf(samples) for samples in zip(*batch)))
    elif isinstance(elem, collections.abc.Sequence):
        transposed = zip(*batch)
        return transposed
    raise TypeError(default_collate_err_msg_format.format(elem_type))



import traceback
import logging
import attr

from .. import entities, exceptions

logger = logging.getLogger(name='dtlpy')


@attr.s
class User(entities.BaseEntity):
    """
    User entity
    """
    created_at = attr.ib()
    updated_at = attr.ib(repr=False)
    name = attr.ib()
    last_name = attr.ib()
    username = attr.ib()
    avatar = attr.ib(repr=False)
    email = attr.ib()
    role = attr.ib()
    type = attr.ib()
    org = attr.ib()
    id = attr.ib()

    # api
    _project = attr.ib(repr=False)
    _client_api = attr.ib(default=None, repr=False)
    _users = attr.ib(repr=False, default=None)

    @property
    def createdAt(self):
        return self.created_at

    @property
    def updatedAt(self):
        return self.updated_at

    @staticmethod
    def _protected_from_json(_json, project, client_api, users=None):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform json
        :param project: project entity
        :param client_api: ApiClient entity
        :param users: Users repository
        :return:
        """
        try:
            user = User.from_json(_json=_json,
                                  project=project,
                                  users=users,
                                  client_api=client_api)
            status = True
        except Exception:
            user = traceback.format_exc()
            status = False
        return status, user

    @property
    def project(self):
        if self._project is None:
            raise exceptions.PlatformException(error='2001',
                                               message='Missing entity "project".')
        assert isinstance(self._project, entities.Project)
        return self._project

    @classmethod
    def from_json(cls, _json, project, client_api, users=None):
        """
        Build a User entity object from a json

        :param dict _json: _json response from host
        :param dtlpy.entities.project.Project project: project entity
        :param client_api: ApiClient entity
        :param users: Users repository
        :return: User object
        :rtype: dtlpy.entities.user.User
        """
        return cls(
            created_at=_json.get('createdAt', None),
            name=_json.get('firstName', None),
            updated_at=_json.get('updatedAt', None),
            last_name=_json.get('lastName', None),
            username=_json.get('username', None),
            avatar=_json.get('avatar', None),
            email=_json.get('email', None),
            role=_json.get('role', None),
            type=_json.get('type', None),
            org=_json.get('org', None),
            id=_json.get('id', None),
            project=project,
            users=users,
            client_api=client_api)

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(self,
                            filter=attr.filters.exclude(attr.fields(User)._project,
                                                        attr.fields(User).name,
                                                        attr.fields(User)._client_api,
                                                        attr.fields(User)._users,
                                                        attr.fields(User).last_name,
                                                        attr.fields(User).created_at,
                                                        attr.fields(User).updated_at,
                                                        ))
        _json['firstName'] = self.name
        _json['lastName'] = self.last_name
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at
        return _json



import os

DATALOOP_PATH = os.environ['DATALOOP_PATH'] if 'DATALOOP_PATH' in os.environ \
    else os.path.join(os.path.expanduser('~'), '.dataloop')
DEFAULT_ENVIRONMENT = 'https://gate.dataloop.ai/api/v1'
DEFAULT_ENVIRONMENTS = {
    'https://dev-gate.dataloop.ai/api/v1':
        {'alias': 'dev',
         'audience': 'https://dataloop-development.auth0.com/api/v2/',
         'client_id': 'I4Arr9ixs5RT4qIjOGtIZ30MVXzEM4w8',
         'auth0_url': 'https://dataloop-development.auth0.com',
         'gate_url': 'https://dev-gate.dataloop.ai',
         'token': None,
         'refresh_token': None,
         'verify_ssl': True,
         'url': 'https://dev-con.dataloop.ai/'},
    'https://rc-gate.dataloop.ai/api/v1':
        {'alias': 'rc',
         'audience': 'https://dataloop-development.auth0.com/api/v2/',
         'client_id': 'I4Arr9ixs5RT4qIjOGtIZ30MVXzEM4w8',
         'auth0_url': 'https://dataloop-development.auth0.com',
         'gate_url': 'https://rc-gate.dataloop.ai',
         'token': None,
         'refresh_token': None,
         'verify_ssl': True,
         'url': 'https://rc-con.dataloop.ai/'},
    'https://custom1-gate.dataloop.ai/api/v1': {
        'gate_url': 'https://custom1-gate.dataloop.ai',
        'auth0_url': 'https://dataloop-development.auth0.com',
        'verify_ssl': True,
        'client_id': 'NEED',
        'audience': 'https://dataloop-development.auth0.com/api/v2/',
        'token': '',
        'refresh_token': '',
        'alias': 'new-dev'},
    'https://gate.dataloop.ai/api/v1': {
        'alias': 'prod',
        'audience': 'https://dataloop-production.auth0.com/userinfo',
        'client_id': 'FrG0HZga1CK5UVUSJJuDkSDqItPieWGW',
        'auth0_url': 'https://dataloop-production.auth0.com',
        'gate_url': 'https://gate.dataloop.ai',
        'token': None,
        'refresh_token': None,
        'verify_ssl': True,
        'url': 'https://console.dataloop.ai/'},
    'https://localhost:8443/api/v1': {
        'alias': 'local',
        'audience': 'https://dataloop-local.auth0.com/userinfo',
        'client_id': 'ewGhbg5brMHOoL2XZLHBzhEanapBIiVO',
        'auth0_url': 'https://dataloop-local.auth0.com',
        'gate_url': 'https://localhost:8443',
        'token': None,
        'refresh_token': None,
        'verify_ssl': False},
    'https://poc-gate.dataloop.ai/api/v1': {
        'alias': 'aws',
        'audience': 'https://dataloop-aws-poc.eu.auth0.com/api/v2/',
        'client_id': 'dHHctbFPa4TFgo1hh9Ig2Fyh71N46BEM',
        'auth0_url': 'https://dataloop-aws-poc.eu.auth0.com',
        'gate_url': 'https://poc-gate.dataloop.ai',
        'token': None,
        'refresh_token': None,
        'verify_ssl': True},
    'https://host.docker.internal:8443/api/v1': {
        'alias': 'docker_windows',
        'audience': 'https://dataloop-local.auth0.com/userinfo',
        'client_id': 'ewGhbg5brMHOoL2XZLHBzhEanapBIiVO',
        'auth0_url': 'https://dataloop-local.auth0.com',
        'gate_url': 'https://172.17.0.1:8443',
        'token': None,
        'refresh_token': None,
        'verify_ssl': False},
    'https://docker.for.mac.localhost:8443/api/v1': {
        'alias': 'minikube_local_mac',
        'audience': 'https://dataloop-local.auth0.com/userinfo',
        'client_id': 'ewGhbg5brMHOoL2XZLHBzhEanapBIiVO',
        'auth0_url': 'https://dataloop-local.auth0.com',
        'gate_url': 'https://host.docker.internal:8443',
        'token': None,
        'refresh_token': None,
        'verify_ssl': False},
    'http://kong/api/v1': {
        'alias': 'minikube',
        'audience': 'https://dataloop-local.auth0.com/userinfo',
        'client_id': 'ewGhbg5brMHOoL2XZLHBzhEanapBIiVO',
        'auth0_url': 'https://dataloop-local.auth0.com',
        'gate_url': 'http://kong',
        'token': None,
        'refresh_token': None,
        'verify_ssl': False}
}



import json
from typing import List
import logging
import os

logger = logging.getLogger(name='dtlpy')


class Layer:
    def __init__(self, name, layer_type, url):
        self.name = name
        self.type = layer_type
        self.url = url


class ItemGis:
    def __init__(self,
                 name: str,
                 data: dict = None,
                 layer: Layer = None,
                 optional_layers: List[Layer] = None,
                 zoom: int = None,
                 min_zoom: int = None,
                 max_zoom: int = None,
                 epsg: str = None,
                 bounds: list = None,
                 aoi: list = None):
        self.name = name

        self.layer = layer or Layer(name=data.get('name', None), layer_type=data.get('type', None),
                                    url=data.get('url', None))
        if self.layer is None:
            raise ValueError('layer is required')
        elif self.layer is not None and isinstance(self.layer, dict):
            self.layer = Layer(name=self.layer.get('name', None), layer_type=self.layer.get('type', None), url=self.layer.get('url', None))


        self.optional_layers = optional_layers or [
            Layer(name=layer.get('name', None), layer_type=layer.get('type', None), url=layer.get('url', None)) for
            layer in data.get('optionalLayers', [])]

        if self.optional_layers is not None and isinstance(optional_layers, list):
            new_optional_layers = []
            for op_layer in self.optional_layers:
                if isinstance(op_layer, dict):
                    new_optional_layers.append(Layer(name=op_layer.get('name', None), layer_type=op_layer.get('type', None), url=op_layer.get('url', None)))
                else:
                    new_optional_layers.append(op_layer)
            self.optional_layers = new_optional_layers

        self.epsg = epsg or data.get('epsg', None)
        if self.epsg is None:
            raise ValueError('epsg is required')

        self.zoom = zoom or data.get('zoom', None)
        self.min_zoom = min_zoom or data.get('minZoom', None)
        self.max_zoom = max_zoom or data.get('maxZoom', None)
        self.bounds = bounds or data.get('bounds', None)
        self.aoi = aoi or data.get('aoi', None)

    def to_json(self):
        _json = {
            "type": "gis",
            "shebang": "dataloop",
            "metadata": {
                "dltype": "gis"
            },
            'layer': {
                'name': self.layer.name,
                'type': self.layer.type,
                'url': self.layer.url
            },
            "epsg": self.epsg
        }
        if self.optional_layers is not None:
            _json['optionalLayers'] = [
                {
                    'name': layer.name,
                    'type': layer.type,
                    'url': layer.url
                } for layer in self.optional_layers
            ]
        if self.zoom is not None:
            _json['zoom'] = self.zoom
        if self.min_zoom is not None:
            _json['minZoom'] = self.min_zoom
        if self.max_zoom is not None:
            _json['maxZoom'] = self.max_zoom
        if self.bounds is not None:
            _json['bounds'] = self.bounds
        if self.aoi is not None:
            _json['aoi'] = self.aoi
        return _json

    @classmethod
    def from_local_file(cls, filepath):
        """
        Create a new prompt item from a file
        :param filepath: path to the file
        :return: PromptItem object
        """
        if os.path.exists(filepath) is False:
            raise FileNotFoundError(f'File does not exists: {filepath}')
        if 'json' not in os.path.splitext(filepath)[-1]:
            raise ValueError(f'Expected path to json item, got {os.path.splitext(filepath)[-1]}')
        with open(filepath, 'r', encoding='utf-8') as f:
            data = json.load(f)
        return cls(name=os.path.basename(filepath), data=data)


import sys

import pandas as pd
import numpy as np
import traceback
import datetime
import logging
import tqdm
import uuid
import os

from .. import entities
from . import BaseModelAdapter, metrics

logger = logging.getLogger(name='dtlpy')


def mean_or_nan(arr):
    if isinstance(arr, list) and len(arr) == 0:
        return np.nan
    else:
        return np.mean(arr)


def model_info_name(model: entities.Model, package: entities.Package):
    if model is None:
        return "{}-no-model".format(package.name)
    else:
        return "{}-{}".format(package.name, model.name)


def measure_annotations(
        annotations_set_one: entities.AnnotationCollection,
        annotations_set_two: entities.AnnotationCollection,
        match_threshold=0.5,
        ignore_labels=False,
        ignore_attributes=False,
        compare_types=None):
    """
    Compares list (or collections) of annotations
    This will also return the precision and recall of the two sets, given that the first that is a GT and the second set
    is the detection (this affects the denominator of the calculation).


    :param annotations_set_one: dl.AnnotationCollection entity with a list of annotations to compare
    :param annotations_set_two: dl.AnnotationCollection entity with a list of annotations to compare
    :param match_threshold: IoU threshold to count as a match
    :param ignore_labels: ignore label when comparing - measure only geometry
    :param ignore_attributes: ignore attribute score for final annotation score
    :param compare_types: list of type to compare. enum dl.AnnotationType

    Returns a dictionary of all the compare data
    """

    if compare_types is None:
        compare_types = [entities.AnnotationType.BOX,
                         entities.AnnotationType.CLASSIFICATION,
                         entities.AnnotationType.POLYGON,
                         entities.AnnotationType.POINT,
                         entities.AnnotationType.SEGMENTATION]
    final_results = dict()
    all_scores = list()
    true_positives = 0
    false_positives = 0
    false_negatives = 0

    # for local annotations - set random id if None
    for annotation in annotations_set_one:
        if annotation.id is None:
            annotation.id = str(uuid.uuid1())
    for annotation in annotations_set_two:
        if annotation.id is None:
            annotation.id = str(uuid.uuid1())

    # start comparing
    for compare_type in compare_types:
        matches = metrics.Matches()
        annotation_subset_one = entities.AnnotationCollection()
        annotation_subset_two = entities.AnnotationCollection()
        annotation_subset_one.annotations = [a for a in annotations_set_one if
                                             a.type == compare_type and not a.metadata.get('system', dict()).get(
                                                 'system', False)]
        annotation_subset_two.annotations = [a for a in annotations_set_two if
                                             a.type == compare_type and not a.metadata.get('system', dict()).get(
                                                 'system', False)]
        # create 2d dataframe with annotation id as names and set all to -1 -> not calculated
        if ignore_labels:
            matches = metrics.Matchers.general_match(matches=matches,
                                                     first_set=annotation_subset_one,
                                                     second_set=annotation_subset_two,
                                                     match_type=compare_type,
                                                     match_threshold=match_threshold,
                                                     ignore_labels=ignore_labels,
                                                     ignore_attributes=ignore_attributes)
        else:
            unique_labels = np.unique([a.label for a in annotation_subset_one] +
                                      [a.label for a in annotation_subset_two])
            for label in unique_labels:
                first_set = [a for a in annotation_subset_one if a.label == label]
                second_set = [a for a in annotation_subset_two if a.label == label]
                matches = metrics.Matchers.general_match(matches=matches,
                                                         first_set=first_set,
                                                         second_set=second_set,
                                                         match_type=compare_type,
                                                         match_threshold=match_threshold,
                                                         ignore_labels=ignore_labels,
                                                         ignore_attributes=ignore_attributes
                                                         )
        if len(matches) == 0:
            continue
        all_scores.extend(matches.to_df()['annotation_score'])
        final_results[compare_type] = metrics.Results(matches=matches,
                                                      annotation_type=compare_type)
        true_positives += final_results[compare_type].summary()['n_annotations_matched_total']
        false_positives += final_results[compare_type].summary()['n_annotations_unmatched_set_two']
        false_negatives += final_results[compare_type].summary()['n_annotations_unmatched_set_one']

    final_results['total_mean_score'] = mean_or_nan(all_scores)
    final_results['precision'] = true_positives / (true_positives + false_positives)
    final_results['recall'] = true_positives / (true_positives + false_negatives)
    return final_results


def measure_item(ref_item: entities.Item, test_item: entities.Item,
                 ref_project: entities.Project = None, test_project: entities.Project = None,
                 ignore_labels=False,
                 ignore_attributes=False,
                 match_threshold=0.5,
                 pbar=None):
    """
    Compare annotations sets between two items

    :param ref_item:
    :param test_item:
    :param ref_project:
    :param test_project:
    :param ignore_labels:
    :param ignore_attributes:
    :param match_threshold:
    :param pbar:
    :return:
    """
    try:
        annotations_set_one = ref_item.annotations.list()
        annotations_set_two = test_item.annotations.list()
        final = measure_annotations(annotations_set_one=annotations_set_one,
                                    annotations_set_two=annotations_set_two,
                                    ignore_labels=ignore_labels,
                                    ignore_attributes=ignore_attributes,
                                    match_threshold=match_threshold)

        # get times
        try:
            ref_item_duration_s = metrics.item_annotation_duration(item=ref_item, project=ref_project)
            ref_item_duration = str(datetime.timedelta(seconds=int(np.round(ref_item_duration_s))))
        except Exception:
            ref_item_duration_s = -1
            ref_item_duration = ''

        try:
            test_item_duration_s = metrics.item_annotation_duration(item=test_item, project=test_project)
            test_item_duration = str(datetime.timedelta(seconds=int(np.round(test_item_duration_s))))
        except Exception:
            test_item_duration_s = -1
            test_item_duration = ''

        final.update({'ref_url': ref_item.platform_url,
                      'test_url': test_item.platform_url,
                      'filename': ref_item.filename,
                      'ref_item_duration[s]': ref_item_duration_s,
                      'test_item_duration[s]': test_item_duration_s,
                      'diff_duration[s]': test_item_duration_s - ref_item_duration_s,
                      # round to sec
                      'ref_item_duration': ref_item_duration,
                      'test_item_duration': test_item_duration,
                      })

        return True, final
    except Exception:
        fail_msg = 'failed measuring. ref_item: {!r}, test_item: {!r}'.format(ref_item.id, test_item.id)
        return False, '{}\n{}'.format(fail_msg, traceback.format_exc())
    finally:
        if pbar is not None:
            pbar.update()


def measure_items(ref_items, ref_project, ref_dataset, ref_name,
                  test_items, test_project, test_dataset, test_name,
                  dump_path=None):
    from multiprocessing.pool import ThreadPool
    ref_items_filepath_dict = {item.filename: item for page in ref_items for item in page}
    test_items_filepath_dict = {item.filename: item for page in test_items for item in page}
    pool = ThreadPool(processes=32)
    pbar = tqdm.tqdm(total=len(ref_items_filepath_dict), file=sys.stdout)
    jobs = dict()
    for filepath, ref_item in ref_items_filepath_dict.items():
        if filepath in test_items_filepath_dict:
            test_item = test_items_filepath_dict[filepath]
            jobs[ref_item.filename] = pool.apply_async(measure_item, kwds={'test_item': test_item,
                                                                           'ref_item': ref_item,
                                                                           'ref_project': ref_project,
                                                                           'test_project': test_project,
                                                                           'pbar': pbar})
    pool.close()
    pool.join()
    _ = [job.wait() for job in jobs.values()]
    raw_items_summary = dict()
    failed_items_errors = dict()
    for filename, job in jobs.items():
        success, result = job.get()
        if success:
            raw_items_summary[filename] = result
        else:
            failed_items_errors[filename] = result
    pool.terminate()
    pbar.close()

    df, raw_items_summary = create_summary(ref_name=ref_name, test_name=test_name, raw_items_summary=raw_items_summary)

    if len(failed_items_errors) != 0:
        logger.error(failed_items_errors)
    if dump_path is not None:
        save_to_file(dump_path=dump_path,
                     df=df,
                     ref_name=ref_name,
                     test_name=test_name)
    return df, raw_items_summary, failed_items_errors


def create_summary(ref_name, test_name, raw_items_summary):
    summary = list()
    ref_column_name = 'Ref-{!r}'.format(ref_name)
    test_column_name = 'Test-{!r}'.format(test_name)
    for filename, scores in raw_items_summary.items():
        line = {'filename': scores['filename'],
                ref_column_name: scores['ref_url'],
                test_column_name: scores['test_url'],
                'total_score': scores['total_mean_score'],
                'ref_duration[s]': scores['ref_item_duration[s]'],
                'test_duration[s]': scores['test_item_duration[s]'],
                'diff_duration[s]': scores['diff_duration[s]']}
        for tool_type in list(entities.AnnotationType):
            if tool_type in scores:
                res = scores[tool_type].summary()
                line['{}_annotation_score'.format(tool_type)] = res['mean_annotations_scores']
                line['{}_attributes_score'.format(tool_type)] = res['mean_attributes_scores']
                line['{}_ref_number'.format(tool_type)] = res['n_annotations_set_one']
                line['{}_test_number'.format(tool_type)] = res['n_annotations_set_two']
                line['{}_match_number'.format(tool_type)] = res['n_annotations_matched_total']
        summary.append(line)
    df = pd.DataFrame(summary)
    # Drop column only if all the values are None
    df = df.dropna(how='all', axis=1)
    ####

    return df, raw_items_summary


def save_to_file(df, dump_path, ref_name, test_name):
    # df = df.sort_values(by='box_score')
    ref_column_name = 'Ref-{!r}'.format(ref_name)
    test_column_name = 'Test-{!r}'.format(test_name)

    def make_clickable(val):
        return '<a href="{}">{}</a>'.format(val, 'item')

    s = df.style.format({ref_column_name: make_clickable,
                         test_column_name: make_clickable}).render()
    os.makedirs(dump_path, exist_ok=True)
    html_filepath = os.path.join(dump_path, '{}-vs-{}.html'.format(ref_name, test_name))
    csv_filepath = os.path.join(dump_path, '{}-vs-{}.csv'.format(ref_name, test_name))
    with open(html_filepath, 'w') as f:
        f.write(s)
    df.to_csv(csv_filepath)



def main(project_name, dataset_name, item_remote_path):
    """
    Download and show an image with it's annotations
    :return:
    """
    import dtlpy as dl
    import numpy as np
    from PIL import Image

    # Get project and dataset
    project = dl.projects.get(project_name=project_name)
    dataset = project.datasets.get(dataset_name=dataset_name)
    # Get item
    item = dataset.items.get(filepath=item_remote_path)

    # download item as a buffer
    buffer = item.download(save_locally=False)

    # open image
    image = Image.open(buffer)

    # download annotations
    annotations = item.annotations.show(width=image.size[0],
                                        height=image.size[1],
                                        thickness=3)
    annotations = Image.fromarray(annotations.astype(np.uint8))

    # show separate
    annotations.show()
    image.show()

    # plot on top
    image.paste(annotations, (0, 0), annotations)
    image.show()

    #####################################################
    # show annotations from json file (Dataloop format) #
    #####################################################
    # import dtlpy as dl
    # from PIL import Image
    # import json
    #
    # with open('annotations.json', 'r') as f:
    #     data = json.load(f)
    #
    # for annotation in data['annotations']:
    #     annotations = dl.Annotation.from_json(annotation)
    #     mask = annotations.show(width=640,
    #                             height=480,
    #                             thickness=3,
    #                             color=(255, 0, 0))
    #     mask = Image.fromarray(mask.astype(np.uint8))
    #     mask.show()



import json
import sqlite3
import re

from diskcache import Cache
import os
from .base_cache import BaseCache


class DiskCache(BaseCache):
    def __init__(self, name, level=1, options=None, enable_stats=False, ttl=1000):
        if options is None:
            options = dict()
        self.name = name
        self.level = level
        self.ttl = ttl
        self.cache_dir = options.get(
            "cachePath", os.path.join(self.dataloop_path, "cache", name)
        )
        self.cache = Cache(directory=self.cache_dir)
        self.cache.stats(enable=enable_stats)

        self.conn = sqlite3.connect(os.path.join(self.cache_dir, 'cache.db'), check_same_thread=False)
        self.cursor = self.conn.cursor()

    @property
    def dataloop_path(self):
        return os.environ['DATALOOP_PATH'] if 'DATALOOP_PATH' in os.environ \
            else os.path.join(os.path.expanduser('~'), '.dataloop')

    def set(self, key, value):
        """
        set or add a key and value to the cache
        :param key: str or int type of key
        :param value: pickled value
        :return:
        """
        if not isinstance(key, str) and not isinstance(key, int):
            raise ValueError("key must be string or int")
        with Cache(self.cache.directory) as reference:
            reference.set(key=key, value=value, expire=self.ttl)

    def _key_fix(self, key):
        if '**' in key:
            key = key.replace('**', '%')
        if '*' in key:
            key = key.replace('*', '%')
        return key

    def list(self, pattern):
        """
        list the keys py pattern from the cache

        :param pattern: str or int type of key
        :return: the value of the key
        """
        self.cache.close()

        key = self._key_fix(pattern)

        try:
            self.cursor.execute("SELECT key FROM Cache WHERE key LIKE '{}'".format(key))

            rows = self.cursor.fetchall()
            for row in rows:
                for key in row:
                    yield key
        except:
            return None

    def ping(self):
        """
        Cache ping check if connection is working
        """
        if os.path.exists(os.path.join(self.cache_dir, 'cache.db')):
            return True
        else:
            raise Exception('cache db not fond')

    def get(self, key):
        """
        get the value of the key from the cache
        :param key: str or int type of key
        :return: the value of the key
        """
        self.cache.close()
        res = self.cache.get(key=key)
        if res is not None:
            try:
                return json.loads(res)
            except:
                return res

    def delete(self, key):
        """
        delete the element from the cache
        :param key: str or int type of key
        :return:
        """
        self.cache.close()
        self.cache.delete(key=key)

    def add(self, key, value):
        """
        add the element to the cache if the key is not already present.
        :param key: str or int type of key
        :param value: pickled value
        :return: bool: True if add False if not
        """
        if not isinstance(key, str) and not isinstance(key, int):
            raise ValueError("key must be string or int")
        self.cache.close()
        return self.cache.add(key=key, value=value, expire=self.ttl)

    def push(self, value):
        """
        push the element to the cache
        :param value: pickled value
        :return: bool: True if add False if not
        """
        self.cache.close()
        return self.cache.push(value=value, expire=self.ttl)

    def incr(self, key, value=1):
        """
        incremented the value of this key
        :param key: str or int type of key
        :param value: the amount of incremented
        :return:
        """
        self.cache.incr(key=key, delta=value)

    def decr(self, key, value=1):
        """
        decremented  the value of this key
        :param key: str or int type of key
        :param value: the amount of decremented
        :return:
        """
        self.cache.decr(key=key, delta=value)

    def pop(self, key):
        """
        delete the element from the cache and return it
        :param key: str or int type of key
        :return: the deleted element
        """
        self.cache.close()
        return self.cache.pop(key)

    def volume(self):
        """
        returns the estimated total size in bytes of the cache directory on disk.
        :return: size in bytes
        """
        return self.cache.volume()

    def clear(self):
        """
        simply removes all items from the cache.
        :return: number of the items
        """
        return self.cache.clear()

    def stats(self):
        """
        returns cache hits and misses.
        :return: tuple (hits, misses)
        """
        hits, misses = self.cache.stats(enable=False, reset=False)
        self.cache.stats(enable=True)
        return hits, misses

    def reset_stats(self):
        """
        reset cache hits and misses.
        :return:
        """
        self.cache.stats(enable=False, reset=True)
        self.cache.stats(enable=True)

    def keys(self):
        """
        return all the cache keys
        :return: list of the keys
        """
        for output in list(self.cache.iterkeys()):
            if output is not None:
                yield output

    def delete_cache(self):
        """
        Delete the cache folder
        """
        self.cache.close()
        import shutil

        try:
            shutil.rmtree(self.cache.directory)
        except OSError:  # Windows wonkiness
            pass



#! /usr/bin/env python3
# This file is part of DTLPY.
#
# DTLPY is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DTLPY is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DTLPY.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    readme = f.read()

with open('LICENSE') as f:
    license_ = f.read()

with open('requirements.txt') as f:
    requirements = f.read()

setup(name='dtlpy',
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: 3.12',
          'Programming Language :: Python :: 3.13',
          'Programming Language :: Python :: 3.14',
      ],
      version='1.118.13',
      description='SDK and CLI for Dataloop platform',
      author='Dataloop Team',
      author_email='info@dataloop.ai',
      url='https://github.com/dataloop-ai/dtlpy',
      license='Apache License 2.0',
      long_description=readme,
      long_description_content_type='text/markdown',
      packages=find_packages(exclude=('tests', 'docs', 'samples', 'venv*', '.venv*', 'env*', '.env*', 'build*', 'dist*', '*.egg-info*')),
      setup_requires=['wheel'],
      install_requires=requirements,
      test_suite='tests',
      python_requires='>=3.8',
      scripts=['dtlpy/dlp/dlp.py', 'dtlpy/dlp/dlp.bat', 'dtlpy/dlp/dlp'],
      include_package_data=True,
      entry_points={
          'console_scripts': [
              'dlp=dtlpy.dlp.dlp:main',
          ],
      },
      )
